# pylint: disable=W0621,C0111
import json
from decouple import config
import pytest
import requests_mock as rm
from bitfinex.rest import ClientV1 as Client


API_KEY = "test"  # os.environ.get('API_KEY', config('API_KEY'))
API_SECRET = "test"  # os.environ.get('API_SECRET', config('API_SECRET'))

@pytest.fixture()
def client():
    return Client(key="testing", secret="testing")


def test_should_have_server(client):
    assert client.server() == "https://api.bitfinex.com/v1"


def test_should_have_url_for_foo(client):
    assert client.url_for("foo") == "https://api.bitfinex.com/v1/foo"


def test_should_have_url_for_path_arg(client):
    assert client.url_for('foo/%s', path_arg="bar") == (
        "https://api.bitfinex.com/v1/foo/bar"
    )


def test_should_have_url_with_parameters(client):
    assert client.url_for('foo', parameters={'a': 1, 'b': 2}) == (
        "https://api.bitfinex.com/v1/foo?a=1&b=2"
    )

def test_should_have_url_for(client):
    assert client.url_for("foo") == "https://api.bitfinex.com/v1/foo"


def test_should_have_url_for_with_path_arg(client):
    expected = "https://api.bitfinex.com/v1/foo/bar"
    path = "foo/%s"
    assert client.url_for(path, path_arg='bar') == expected
    assert client.url_for(path, 'bar') == expected


def test_should_have_url_for_with_parameters(client):
    expected = "https://api.bitfinex.com/v1/foo?a=1"
    assert client.url_for("foo", parameters={'a': 1}) == expected
    assert client.url_for("foo", None, {'a': 1}) == expected


def test_should_have_url_for_with_path_arg_and_parameters(client):
    expected = "https://api.bitfinex.com/v1/foo/bar?a=1"
    path = "foo/%s"
    assert client.url_for(path, path_arg='bar', parameters={'a': 1}) == expected
    assert client.url_for(path, 'bar', {'a': 1}) == expected


def test_should_have_symbols(client, requests_mock):
    request_result = ["btcusd", "ltcusd", "ltcbtc"]
    requests_mock.register_uri(
        rm.ANY,
        rm.ANY,
        text='["btcusd", "ltcusd", "ltcbtc"]'
    )
    assert client.symbols() == request_result
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v1/symbols'
    )


def test_should_have_symbol_details(client, requests_mock):
    # mock out the request
    request_result = '''[{
        "pair":"btcusd",
        "price_precision":5,
        "initial_margin":"30.0",
        "minimum_margin":"15.0",
        "maximum_order_size":"2000.0",
        "minimum_order_size":"0.01",
        "expiration":"NA"
        },{
        "pair":"ltcusd",
        "price_precision":5,
        "initial_margin":"30.0",
        "minimum_margin":"15.0",
        "maximum_order_size":"5000.0",
        "minimum_order_size":"0.1",
        "expiration":"NA"
        }]'''
    requests_mock.register_uri(
        rm.ANY,
        rm.ANY,
        text=request_result
    )
    result = client.symbols_details()
    assert result == json.loads(request_result)
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v1/symbols_details'
    )


def test_should_have_ticker(client, requests_mock):
    # mock out the request
    request_result = {
        "mid": "562.56495",
        "bid": "562.15",
        "ask": "562.9799",
        "last_price": "562.25",
        "volume": "7842.11542563",
        "timestamp": "1395552658.339936691"
    }
    requests_mock.register_uri(
        rm.ANY,
        rm.ANY,
        text=json.dumps(request_result)
    )
    assert client.ticker('btcusd') == request_result
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v1/pubticker/btcusd'
    )


def test_should_have_today(client, requests_mock):
    # mock out the request
    request_result = {"low": "550.09", "high": "572.2398", "volume": "7305.33119836"}
    requests_mock.register_uri(
        rm.ANY,
        rm.ANY,
        text=json.dumps(request_result)
    )

    assert client.today('btcusd') == request_result
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v1/today/btcusd'
    )


def test_should_have_stats(client, requests_mock):
    request_result = [
        {"period": 1, "volume": "7410.27250155"},
        {"period": 7, "volume": "52251.37118006"},
        {"period": 30, "volume": "464505.07753251"}
    ]
    requests_mock.register_uri(
        rm.ANY,
        rm.ANY,
        text=json.dumps(request_result)
    )
    assert client.stats('btcusd') == request_result
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v1/stats/btcusd'
    )


def test_should_have_lendbook(client, requests_mock):
    request_result = {
        "bids": [
            {"rate": "5.475", "amount": "15.03894663", "period": 30,
             "timestamp": "1395112149.0", "frr": False},
            {"rate": "2.409", "amount": "14.5121868", "period": 7,
             "timestamp": "1395497599.0", "frr": False}
        ],
        "asks": [
            {"rate": "6.351", "amount": "15.5180735", "period": 5,
             "timestamp": "1395549996.0", "frr": False},
            {"rate": "6.3588", "amount": "626.94808249", "period": 30,
             "timestamp": "1395400654.0", "frr": False}
        ]
    }
    requests_mock.register_uri(
        rm.ANY,
        rm.ANY,
        text=json.dumps(request_result)
    )
    assert client.lendbook('btc') == request_result
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v1/lendbook/btc'
    )


def test_should_have_lendbook_with_parameters(client, requests_mock):
    request_result = '{"bids":[{"rate":"5.475","amount":"15.03894663","period":30,\
        "timestamp":"1395112149.0","frr":"No"},{"rate":"2.409",\
        "amount":"14.5121868","period":7,"timestamp":"1395497599.0",\
        "frr":"No"}],"asks":[]}'
    parameters = {'limit_bids': 2, 'limit_asks': 0}
    expected = {
        "bids": [
            {"rate": "5.475", "amount": "15.03894663", "period": 30,
             "timestamp": "1395112149.0", "frr": False},
            {"rate": "2.409", "amount": "14.5121868", "period": 7,
             "timestamp": "1395497599.0", "frr": False}
        ],
        "asks": []
    }
    requests_mock.register_uri(
        rm.ANY,
        rm.ANY,
        text=request_result
    )
    assert client.lendbook('btc', parameters) == expected
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v1/lendbook/btc?limit_asks=0&limit_bids=2'
    )


def test_should_have_order_book(client, requests_mock):
    request_result = {
        "bids": [
            {"price": "562.2601", "amount": "0.985", "timestamp": "1395567556.0"}
        ],
        "asks": [
            {"price": "563.001", "amount": "0.3", "timestamp": "1395532200.0"}
        ]
    }
    requests_mock.register_uri(
        rm.ANY,
        rm.ANY,
        text=json.dumps(request_result)
    )
    assert client.order_book('btcusd') == request_result
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v1/book/btcusd'
    )


def test_should_have_order_book_with_parameters(client, requests_mock):
    parameters = {'limit_asks': 0}
    request_result = {
        "bids": [
            {"price": "562.2601", "amount": "0.985", "timestamp": "1395567556.0"}
        ],
        "asks": []
    }

    requests_mock.register_uri(
        rm.ANY,
        rm.ANY,
        text=json.dumps(request_result)
    )
    assert client.order_book('btcusd', parameters) == request_result
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v1/book/btcusd?limit_asks=0'
    )


def test_get_active_orders_returns_json(client, requests_mock):
    request_result = [{
        "price":"562.2601",
        "amount":"0.985",
        "timestamp":"1395567556.0"
    }]
    requests_mock.register_uri(
        rm.ANY,
        rm.ANY,
        text=json.dumps(request_result)
    )
    assert client.active_orders() == request_result
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v1/orders'
    )


def test_get_active_positions_returns_json(client, requests_mock):
    request_result = [{
        "price":"562.2601",
        "amount":"0.985",
        "timestamp":"1395567556.0"
    }]
    requests_mock.register_uri(
        rm.ANY,
        rm.ANY,
        text=json.dumps(request_result)
    )
    assert client.active_positions() == request_result
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v1/positions'
    )


def test_key_permissions_returns_json(client, requests_mock):
    request_result = {
        'account': {'read': True, 'write': False},
        'history': {'read': True, 'write': False},
        'orders': {'read': True, 'write': False},
        'positions': {'read': True, 'write': False},
        'funding': {'read': True, 'write': False},
        'wallets': {'read': True, 'write': False},
        'withdraw': {'read': False, 'write': False}
    }

    requests_mock.register_uri(
        rm.ANY,
        rm.ANY,
        text=json.dumps(request_result)
    )

    assert client.key_permissions() == request_result
    assert requests_mock.request_history[0].url == (
        'https://api.bitfinex.com/v1/key_info'
    )
# def test_get_full_history(client, requests_mock):
#     mock_body = [{
#         "price":"562.2601",
#         "amount":"0.985",
#         "timestamp":"1395567556.0"
#     }]
#     url = client.url_for(path="orders")
#     httpretty.register_uri(httpretty.POST, url, body=mock_body, status=200)
#
#     ap = client.active_positions()
#     self.assertIsInstance(ap, list)



import pytest
from bitfinex import utils

def test_order_symbol_adds_t_to_symbol():
    assert utils.order_symbol("BTCUSD") == "tBTCUSD"

def test_order_symbol_keeps_t_to_symbol():
    assert utils.order_symbol("tBTCUSD") == "tBTCUSD"

def test_order_symbol_adds_t_to_symbol_lowercase():
    assert utils.order_symbol("btcusd", capital=False) == "tbtcusd"

def test_order_symbol_adds_f_to_symbol():
    assert utils.order_symbol("BTC") == "fBTC"

def test_order_symbol_keeps_f_to_symbol():
    assert utils.order_symbol("fBTC") == "fBTC"

def test_order_symbol_adds_f_to_symbol_lowercase():
    assert utils.order_symbol("btc", capital=False) == "fbtc"

def test_order_symbol_capital_letters():
    assert utils.order_symbol("btcusd") == "tBTCUSD"

def test_order_symbol_passes_on_unknown_symbols_unchanged():
    assert utils.order_symbol("custom_sym") == "custom_sym"



"""Module for websocket utilities"""

ERROR_CODES = {
    10000: "Unknown error",
    10001: "Generic error",
    10008: "Concurrency error",
    10020: "Request parameters error",
    10050: "Configuration setup failed",
    10100: "Failed authentication",
    10111: "Error in authentication request payload",
    10112: "Error in authentication request signature",
    10113: "Error in authentication request encryption",
    10114: "Error in authentication request nonce",
    10200: "Error in un-authentication request",
    10300: "Failed channel subscription",
    10301: "Failed channel subscription: already subscribed",
    10400: "Failed channel un-subscription: channel not found",
    11000: "Not ready, try again later",
    20051: "Websocket server stopping... please reconnect later",
    20060: "Websocket server resyncing... please reconnect later",
    20061: "Websocket server resync complete. please reconnect",
    5000: "Info message",
}


# Abbreviation Glossary
# https://bitfinex.readme.io/v2/docs/abbreviations-glossary#section-abbreviation-glossary
NOTIFICATION_CODES = {
    "bu": "balance update",
    "ps": "position snapshot",
    "pn": "new position",
    "pu": "position update",
    "pc": "position close",
    "ws": "wallet snapshot",
    "wu": "wallet update",
    "os": "order snapshot",
    "on": "order new",
    "ou": "order update",
    "oc": "order cancel",
    "ox_multi": "order multi-op",
    "oc-req": "order cancel request",
    "oc_multi-req": "multiple orders cancel request",
    "te": "trade executed",
    "tu": "trade execution update",
    "fte": "funding trade execution",
    "ftu": "funding trade update",
    "hos": "historical order snapshot",
    "mis": "margin information snapshot",
    "miu": "margin information update",
    "n": "notification",
    "fos": "funding offer snapshot",
    "fon": "funding offer new",
    "fou": "funding offer update",
    "foc": "funding offer cancel",
    "hfos": "historical funding offer snapshot",
    "fcs": "funding credits snapshot",
    "fcn": "funding credits new",
    "fcu": "funding credits update",
    "fcc": "funding credits close",
    "hfcs": "historical funding credits snapshot",
    "fls": "funding loan snapshot",
    "fln": "funding loan new",
    "flu": "funding loan update",
    "flc": "funding loan close",
    "hfls": "historical funding loan snapshot",
    "hfts": "historical funding trade snapshot",
    "uac": "user custom price alert",
}


def get_notification_code(description):
    index = list(NOTIFICATION_CODES.values()).index(description)
    return list(NOTIFICATION_CODES.keys())[index]


ORDER_TYPES = [
    "MARKET",
    "EXCHANGE MARKET",
    "LIMIT",
    "EXCHANGE LIMIT",
    "STOP",
    "EXCHANGE STOP",
    "TRAILING STOP",
    "EXCHANGE TRAILING STOP",
    "FOK",
    "EXCHANGE FOK",
    "STOP LIMIT",
    "EXCHANGE STOP LIMIT",
]



# coding=utf-8
import threading
import json
import hmac
import hashlib
from autobahn.twisted.websocket import WebSocketClientFactory, \
    WebSocketClientProtocol, \
    connectWS
from twisted.internet import reactor, ssl
from twisted.internet.protocol import ReconnectingClientFactory
from twisted.internet.error import ReactorAlreadyRunning
from bitfinex import utils
from . import abbreviations

# Example used to make send logic
# https://stackoverflow.com/questions/18899515/writing-an-interactive-client-with-twisted-autobahn-websockets


class BitfinexClientProtocol(WebSocketClientProtocol):

    def __init__(self, factory, payload=None):
        super().__init__()
        self.factory = factory
        self.payload = payload

    def onOpen(self):
        self.factory.protocol_instance = self

    def onConnect(self, response):
        if self.payload:
            self.sendMessage(self.payload, isBinary=False)
        # reset the delay after reconnecting
        self.factory.resetDelay()

    def onMessage(self, payload, isBinary):
        if not isBinary:
            try:
                payload_obj = json.loads(payload.decode('utf8'))
            except ValueError:
                pass
            else:
                self.factory.callback(payload_obj)


class BitfinexReconnectingClientFactory(ReconnectingClientFactory):

    # set initial delay to a short time
    initialDelay = 0.1

    maxDelay = 20

    maxRetries = 30


class BitfinexClientFactory(WebSocketClientFactory, BitfinexReconnectingClientFactory):

    def __init__(self, *args, payload=None, **kwargs):
        WebSocketClientFactory.__init__(self, *args, **kwargs)
        self.protocol_instance = None
        self.base_client = None
        self.payload = payload

    protocol = BitfinexClientProtocol
    _reconnect_error_payload = {
        'e': 'error',
        'm': 'Max reconnect retries reached'
    }

    def clientConnectionFailed(self, connector, reason):
        self.retry(connector)
        if self.retries > self.maxRetries:
            self.callback(self._reconnect_error_payload)

    def clientConnectionLost(self, connector, reason):
        self.retry(connector)
        if self.retries > self.maxRetries:
            self.callback(self._reconnect_error_payload)

    def buildProtocol(self, addr):
        return BitfinexClientProtocol(self, payload=self.payload)


class BitfinexSocketManager(threading.Thread):

    STREAM_URL = 'wss://api.bitfinex.com/ws/2'

    def __init__(self):  # client
        """Initialise the BitfinexSocketManager"""
        threading.Thread.__init__(self)
        self.factories = {}
        self._connected_event = threading.Event()
        self._conns = {}
        self._user_timer = None
        self._user_listen_key = None
        self._user_callback = None

    def _start_socket(self, id_, payload, callback):
        if id_ in self._conns:
            return False

        factory_url = self.STREAM_URL
        factory = BitfinexClientFactory(factory_url, payload=payload)
        factory.base_client = self
        factory.protocol = BitfinexClientProtocol
        factory.callback = callback
        factory.reconnect = True
        self.factories[id_] = factory
        reactor.callFromThread(self.add_connection, id_)

    def add_connection(self, id_):
        """
        Convenience function to connect and store the resulting
        connector.
        """
        factory = self.factories[id_]
        context_factory = ssl.ClientContextFactory()
        self._conns[id_] = connectWS(factory, context_factory)

    def stop_socket(self, conn_key):
        """Stop a websocket given the connection key

        Parameters
        ----------
        conn_key : str
            Socket connection key

        Returns
        -------
        str, bool
            connection key string if successful, False otherwise
        """
        if conn_key not in self._conns:
            return

        # disable reconnecting if we are closing
        self._conns[conn_key].factory = WebSocketClientFactory(self.STREAM_URL)
        self._conns[conn_key].disconnect()
        del self._conns[conn_key]

    def run(self):
        try:
            reactor.run(installSignalHandlers=False)
        except ReactorAlreadyRunning:
            # Ignore error about reactor already running
            pass

    def close(self):
        """Close all connections
        """
        keys = set(self._conns.keys())
        for key in keys:
            self.stop_socket(key)
        self._conns = {}


class WssClient(BitfinexSocketManager):
    """Websocket client for bitfinex.

    Parameters
    ----------
    key : str
        Your API key.

    secret : str
        Your API secret


    .. Hint::

        Do not store your key or secret directly in the code.
        Use environment variables. and fetch them with
        ``os.environ.get("BITFINEX_KEY")``

    """

    ###########################################################################
    # Bitfinex commands
    ###########################################################################

    def __init__(self, key=None, secret=None, nonce_multiplier=1.0):  # client
        super().__init__()
        self.key = key
        self.secret = secret
        self.nonce_multiplier = nonce_multiplier

    def stop(self):
        """Tries to close all connections and finally stops the reactor.
        Properly stops the program."""
        try:
            self.close()
        finally:
            reactor.stop()

    def _nonce(self):
        """Returns a nonce used in authentication.
        Nonce must be an increasing number, if the API key has been used
        earlier or other frameworks that have used higher numbers you might
        need to increase the nonce_multiplier."""
        return str(utils.get_nonce(self.nonce_multiplier))

    def authenticate(self, callback, filters=None):
        """Method used to create an authenticated channel that both recieves
        account spesific messages and is used to send account spesific messages.
        So in order to be able to use the new_order method, you have to
        create a authenticated channel before starting the connection.

        Parameters
        ----------
        callback : func
            A function to use to handle incomming messages. This channel wil
            be handling all messages returned from operations like new_order or
            cancel_order, so make sure you handle all these messages.

        filters : List[str]
            A list of filter strings. See more information here:
            https://docs.bitfinex.com/v2/docs/ws-auth#section-channel-filters

        Example
        -------
         ::

            def handle_account_messages(message):
                print(message)

             # You should only need to create and authenticate a client once.
             # Then simply reuse it later
             my_client = WssClient(key, secret)
             my_client.authenticate(
                callback=handle_account_messages
             )
             my_client.start()

        """
        nonce = self._nonce()
        auth_payload = 'AUTH{}'.format(nonce)
        signature = hmac.new(
            self.secret.encode(),  # settings.API_SECRET.encode()
            msg=auth_payload.encode('utf8'),
            digestmod=hashlib.sha384
        ).hexdigest()
        data = {
            # docs: http://bit.ly/2CEx9bM
            'event': 'auth',
            'apiKey': self.key,
            'authSig': signature,
            'authPayload': auth_payload,
            'authNonce': nonce,
            'calc': 1
        }
        if filters:
            data['filter'] = filters
        payload = json.dumps(data, ensure_ascii=False).encode('utf8')
        return self._start_socket("auth", payload, callback)

    def subscribe_to_ticker(self, symbol, callback):
        """Subscribe to the passed symbol ticks data channel.

        Parameters
        ----------
        symbol : str
            Symbol to request data for.

        callback : func
            A function to use to handle incomming messages

        Example
        -------
         ::

            def my_handler(message):
                # Here you can do stuff with the messages
                print(message)

            # You should only need to create and authenticate a client once.
            # Then simply reuse it later
            my_client = WssClient(key, secret)
            my_client.authenticate(print)
            my_client.subscribe_to_ticker(
                symbol="BTCUSD",
                callback=my_handler
            )
            my_client.start()
        """
        symbol = utils.order_symbol(symbol)
        id_ = "_".join(["ticker", symbol])
        data = {
            'event': 'subscribe',
            'channel': 'ticker',
            'symbol': symbol,
        }
        payload = json.dumps(data, ensure_ascii=False).encode('utf8')
        return self._start_socket(id_, payload, callback)

    def subscribe_to_trades(self, symbol, callback):
        """Subscribe to the passed symbol trades data channel.

        Parameters
        ----------
        symbol : str
            Symbol to request data for.

        callback : func
            A function to use to handle incomming messages

        Example
        -------
         ::

            def my_handler(message):
                # Here you can do stuff with the messages
                print(message)

            # You should only need to create and authenticate a client once.
            # Then simply reuse it later
            my_client = WssClient(key, secret)
            my_client.authenticate(print)
            my_client.subscribe_to_trades(
                symbol="BTCUSD",
                callback=my_handler
            )
            my_client.start()
        """
        symbol = utils.order_symbol(symbol)
        id_ = "_".join(["trades", symbol])
        data = {
            'event': 'subscribe',
            'channel': 'trades',
            'symbol': symbol,
        }
        payload = json.dumps(data, ensure_ascii=False).encode('utf8')
        return self._start_socket(id_, payload, callback)

    # Precision: R0, P0, P1, P2, P3
    def subscribe_to_orderbook(self, symbol, precision, callback):
        """Subscribe to the orderbook of a given symbol.

        Parameters
        ----------
        symbol : str
            Symbol to request data for.

        precision : str
            Accepted values as strings {R0, P0, P1, P2, P3}

        callback : func
            A function to use to handle incomming messages

        Example
        -------
         ::

            def my_handler(message):
                # Here you can do stuff with the messages
                print(message)

            # You should only need to create and authenticate a client once.
            # Then simply reuse it later
            my_client = WssClient(key, secret)

            my_client.subscribe_to_orderbook(
                symbol="BTCUSD",
                precision="P1",
                callback=my_handler
            )
            my_client.start()
        """
        symbol = utils.order_symbol(symbol)
        id_ = "_".join(["order", symbol])
        data = {
            'event': 'subscribe',
            "channel": "book",
            "prec": precision,
            'symbol': symbol,
        }
        payload = json.dumps(data, ensure_ascii=False).encode('utf8')
        return self._start_socket(id_, payload, callback)

    def subscribe_to_candles(self, symbol, timeframe, callback):
        """Subscribe to the passed symbol's OHLC data channel.

        Parameters
        ----------
        symbol : str
            Symbol to request data for

        timeframe : str
            Accepted values as strings {1m, 5m, 15m, 30m, 1h, 3h, 6h, 12h,
            1D, 7D, 14D, 1M}

        callback : func
            A function to use to handle incomming messages

        Returns
        -------
        str
            The socket identifier.

        Example
        -------
         ::

            def my_candle_handler(message):
                # Here you can do stuff with the candle bar messages
                print(message)

            # You should only need to create and authenticate a client once.
            # Then simply reuse it later
            my_client = WssClient(key, secret)

            my_client.subscribe_to_candles(
                symbol="BTCUSD",
                timeframe="1m",
                callback=my_candle_handler
            )
            my_client.subscribe_to_candles(
                symbol="ETHUSD",
                timeframe="5m",
                callback=my_candle_handler
            )
            my_client.start()

        """

        valid_tfs = ['1m', '5m', '15m', '30m', '1h', '3h', '6h', '12h', '1D',
                     '7D', '14D', '1M']
        if timeframe:
            if timeframe not in valid_tfs:
                raise ValueError("timeframe must be any of %s" % valid_tfs)
        else:
            timeframe = '1m'
        identifier = ('candles', symbol, timeframe)
        id_ = "_".join(identifier)
        symbol = utils.order_symbol(symbol)
        key = 'trade:' + timeframe + ':' + symbol
        data = {
            'event': 'subscribe',
            'channel': 'candles',
            'key': key,
        }
        payload = json.dumps(data, ensure_ascii=False).encode('utf8')
        return self._start_socket(id_, payload, callback)

    def ping(self, channel="auth"):
        """Ping bitfinex.

        Parameters
        ----------
        channel : str
            What channel id that should be pinged. Default "auth".
            To get channel id’s use ´client._conns.keys()´.
        """
        client_cid = utils.create_cid()
        data = {
            'event': 'ping',
            'cid': client_cid
        }
        payload = json.dumps(data, ensure_ascii=False).encode('utf8')
        self.factories[channel].protocol_instance.sendMessage(payload, isBinary=False)
        return client_cid

    def new_order_op(self, order_type, symbol, amount, price, price_trailing=None,
                     price_aux_limit=None, price_oco_stop=None, hidden=0,
                     flags=None, tif=None, set_cid=True):
        """Create new order operation

        Parameters
        ----------
        order_type : str
            Order type. Must be one of: "LIMIT", "STOP", "MARKET",
            "TRAILING STOP", "FOK", "STOP LIMIT" or equivelent with "EXCHANGE"
            prepended to it. All orders starting with EXCHANGE are made on the
            exchange wallet. Orders without it is made on the margin wallet and
            will start or change a position.

        symbol : str
            The currency symbol to be traded. e.g. BTCUSD

        amount : decimal str
            The amount to be traided.

        price : decimal str
            The price to buy at. Will be ignored for market orders.

        price_trailing : decimal string
            The trailing price

        price_aux_limit : decimal string
            Auxiliary Limit price (for STOP LIMIT)

        price_oco_stop : decimal string
            OCO stop price

        hidden : bool
            Whether or not to use the hidden order type.

        flags : list
            A list of integers for the different flags. Will be added together
            into a unique integer.

        tif : datetime string

        set_cid : bool
            wheter or not to set a cid.

        Returns
        -------
        dict
            A dict containing the order detials. Used in new_order and for
            creating multiorders.

        Example
        -------
        Note if you only want to create a new order, use the ´´new_order´´
        method bellow. However if you want to submitt multiple order and
        cancel orders at the same time use this method to create order
        operations and send them with the ``multi_order`` method::

            # You should only need to create and authenticate a client once.
            # Then simply reuse it later
            my_client = WssClient(key, secret)
            my_client.authenticate()
            my_client.start()

            order_operation = my_client.new_order_op(
                order_type="LIMIT",
                symbol="BTCUSD",
                amount=0.004,
                price=1000.0
            )

            # Usefull to keep track of an order by its client id, for later
            # operations (like cancel order).
            clinet_id = order_operation["cid"]

            my_client.multi_order(
                operations=[order_operation]
            )

        """
        flags = flags or []
        order_op = {
            'type': order_type,
            'symbol': utils.order_symbol(symbol),
            'amount': amount,
            'price': price,
            'hidden': hidden,
            "flags": sum(flags),
        }
        if price_trailing:
            order_op['price_trailing'] = price_trailing

        if price_aux_limit:
            order_op['price_aux_limit'] = price_aux_limit

        if price_oco_stop:
            order_op['price_oco_stop'] = price_oco_stop

        if tif:
            order_op['tif'] = tif

        if set_cid:
            client_order_id = utils.create_cid()
            order_op['cid'] = client_order_id

        order_op = [
            abbreviations.get_notification_code('order new'),
            order_op
        ]
        return order_op

    def new_order(self, order_type, symbol, amount, price, price_trailing=None,
                  price_aux_limit=None, price_oco_stop=None, hidden=0,
                  flags=None, tif=None, set_cid=True):
        """
        Create new order.

        Parameters
        ----------
        order_type : str
            Order type. Must be one of: "LIMIT", "STOP", "MARKET",
            "TRAILING STOP", "FOK", "STOP LIMIT" or equivelent with "EXCHANGE"
            prepended to it. All orders starting with EXCHANGE are made on the
            exchange wallet. Orders without it is made on the margin wallet and
            will start or change a position.

        symbol : str
            The currency symbol to be traded. e.g. BTCUSD

        amount : decimal string
            The amount to be traided.

        price : decimal string
            The price to buy at. Will be ignored for market orders.

        price_trailing : decimal string
            The trailing price

        price_aux_limit : decimal string
            Auxiliary Limit price (for STOP LIMIT)

        price_oco_stop : decimal string
            OCO stop price

        hidden : bool
            Whether or not to use the hidden order type.

        flags : list
            A list of integers for the different flags. Will be added together
            into a unique integer.

        tif : datetime string

        set_cid : bool
            wheter or not to set a cid.

        Returns
        -------
        int
            Order client id (cid). The CID is also a mts date stamp of when the
            order was created.


        Example
        -------
         ::

            # You should only need to create and authenticate a client once.
            # Then simply reuse it later
            my_client = WssClient(key, secret)
            my_client.authenticate()
            my_client.start()

            order_client_id = my_client.new_order(
                order_type="LIMIT",
                symbol="BTCUSD",
                amount=0.004,
                price=1000.0
            )

        """
        operation = self.new_order_op(
            order_type=order_type,
            symbol=symbol,
            amount=amount,
            price=price,
            price_trailing=price_trailing,
            price_aux_limit=price_aux_limit,
            price_oco_stop=price_oco_stop,
            hidden=hidden,
            flags=flags,
            tif=tif,
            set_cid=set_cid
        )
        data = [0, operation[0], None, operation[1]]

        payload = json.dumps(data, ensure_ascii=False).encode('utf8')
        self.factories["auth"].protocol_instance.sendMessage(payload, isBinary=False)
        if set_cid is True:
            return operation[1]["cid"]
        else:
            return None

    def multi_order(self, operations):
        """Multi order operation.

        Parameters
        ----------
        operations : list
            a list of operations. Read more here:
            https://bitfinex.readme.io/v2/reference#ws-input-order-multi-op
            Hint. you can use the self.new_order_op() for easy new order
            operation creation.

        Returns
        -------
        list
            A list of all the client ids created for each order. Returned in
            the order they are given to the method.

        Example
        -------
         ::

            # You should only need to create and authenticate a client once.
            # Then simply reuse it later
            from bitfinex import utils
            my_client = WssClient(key, secret)
            my_client.authenticate()
            my_client.start()

            example_order_cid_to_cancel = 153925861909296

            # docs: http://bit.ly/2BVqwW6
            cancel_order_operation = {
                'cid': example_order_cid_to_cancel,
                'cid_date': utils.cid_to_date(example_order_cid_to_cancel)
            }

            new_order_operation = my_client.new_order_op(
                order_type="LIMIT",
                symbol="BTCUSD",
                amount=0.004,
                price=1000.0
            )

            order_client_id = my_client.multi_order([
                cancel_order_operation,
                new_order_operation
            ])
        """
        data = [
            0,
            abbreviations.get_notification_code('order multi-op'),
            None,
            operations
        ]
        payload = json.dumps(data, ensure_ascii=False).encode('utf8')
        self.factories["auth"].protocol_instance.sendMessage(payload, isBinary=False)
        return [order[1].get("cid", None) for order in operations]

    def cancel_order(self, order_id):
        """Cancel order

        Parameters
        ----------
        order_id : int, str
            Order id created by Bitfinex

        Example
        -------
         ::

            # You should only need to create and authenticate a client once.
            # Then simply reuse it later
            my_client = WssClient(key, secret)
            my_client.authenticate()
            my_client.start()

            my_client.cancel_order(
                order_id=1234
            )

        """
        data = [
            0,
            abbreviations.get_notification_code('order cancel'),
            None,
            {
                # docs: http://bit.ly/2BVqwW6
                'id': order_id
            }
        ]
        payload = json.dumps(data, ensure_ascii=False).encode('utf8')
        self.factories["auth"].protocol_instance.sendMessage(payload, isBinary=False)

    def cancel_order_cid(self, order_cid, order_date):
        """Cancel order using the client id and the date of the cid. Both are
        returned from the new_order command from this library.

        Parameters
        ----------
        order_cid : str
            cid string. e.g. "1234154"

        order_date : str
            Iso formated order date. e.g. "2012-01-23"


        Example
        -------
         ::

            # You should only need to create and authenticate a client once.
            # Then simply reuse it later
            my_client = WssClient(key, secret)
            my_client.authenticate()
            my_client.start()

            # order_cid created by this library is always a milliseconds
            # time stamp. So you can just divide it by 1000 to get the timestamp.
            my_client.cancel_order(
                order_cid=1538911910035,
                order_date=(
                    datetime.utcfromtimestamp(
                        1538911910035/1000.0
                    ).strftime("%Y-%m-%d")
                )
            )

        """
        data = [
            0,
            abbreviations.get_notification_code('order cancel'),
            None,
            {
                # docs: http://bit.ly/2BVqwW6
                'cid': order_cid,
                'cid_date': order_date
            }
        ]
        payload = json.dumps(data, ensure_ascii=False).encode('utf8')
        self.factories["auth"].protocol_instance.sendMessage(payload, isBinary=False)

    def update_order(self, **order_settings):
        """Update order using the order id

        Parameters
        ----------
        id : int64
            Order ID

        gid : int32
            Group Order ID

        price : decimal string
            Price

        amount : decimal string
            Amount

        delta : decimal string
            Change of amount

        price_aux_limit : decimal string
            Auxiliary limit price

        price_trailing : decimal string
            Trailing price delta

        tif : datetime string
            Time-In-Force: datetime for automatic order cancellation (ie. 2020-01-01 10:45:23)
        """
        data = [
            0,
            abbreviations.get_notification_code('order update'),
            None,
            order_settings
        ]
        payload = json.dumps(data, ensure_ascii=False).encode('utf8')
        self.factories["auth"].protocol_instance.sendMessage(payload, isBinary=False)

    def calc(self, *calculations):
        """
        This message will be used by clients to trigger specific calculations,
        so we don't end up in calculating data that is not usually needed.

        You can request calculations to the websocket server that sends you the
        same message, with the required fields.

        List items must be one of the following:

            - margin_sym_SYMBOL (e.g. margin_sym_tBTCUSD)
            - funding_sym_SYMBOL
            - position_SYMBOL
            - wallet_WALLET-TYPE_CURRENCY


        Parameters
        ----------
        *calculations : list
            list of calculations wanted

        Returns
        -------
        None
            Data is returned over the auth channel. See the abbreviation
            glossary: https://docs.bitfinex.com/v2/docs/abbreviations-glossary

        Examples
        --------
         ::

            # You should only need to create and authenticate a client once.
            # Then simply reuse it later
            my_client = WssClient(key, secret)
            my_client.authenticate(print)
            my_client.start()

            my_client.calc(["margin_sym_tBTCUSD", "funding_sym_fUSD"])
            my_client.calc(["margin_sym_tBTCUSD"])
            my_client.calc(["position_tBTCUSD"])
            my_client.calc(["wallet_exachange_USD"])

        .. Note::

            Calculations are on demand, so no more streaming of unnecessary data.
            Websocket server allows up to 30 calculations per batch.
            If the client sends too many concurrent requests (or tries to spam) requests,
            it will receive an error and potentially a disconnection.
            The Websocket server performs a maximum of 8 calculations per second per client.

        """

        data = [
            0,
            'calc',
            None,
            calculations
        ]
        payload = json.dumps(data, ensure_ascii=False).encode('utf8')
        self.factories["auth"].protocol_instance.sendMessage(payload, isBinary=False)



#!/usr/bin/python
#-*- coding: utf-8 -*-

from __future__ import absolute_import
import json
from json.decoder import JSONDecodeError
import base64
import hmac
import hashlib
import requests
from bitfinex import utils

PROTOCOL = "https"
HOST = "api.bitfinex.com"
VERSION = "v1"

PATH_SYMBOLS = "symbols"
PATH_TICKER = "pubticker/%s"
PATH_TODAY = "today/%s"
PATH_STATS = "stats/%s"
PATH_LENDBOOK = "lendbook/%s"
PATH_ORDERBOOK = "book/%s"
PATH_KEYINFO = "key_info"

# HTTP request timeout in seconds
TIMEOUT = 5.0


class BitfinexException(Exception):
    pass


class Client:
    """
    Client for the bitfinex.com API.

    Link for official bitfinex documentation :

    `Bitfinex rest1 docs <https://docs.bitfinex.com/v1/docs>`_

    `Bitfinex rest1 reference <https://docs.bitfinex.com/v1/reference>`_

    Parameters
    ----------
    key : str
        Bitfinex api key

    secret : str
        Bitfinex api secret

    nonce_multiplier : Optional float
        Multiply nonce by this number

    Examples
    --------
     ::

        bfx_client = Client(key,secret)

        bfx_client = Client(key,secret,2.0)
    """

    def __init__(self, key=None, secret=None, nonce_multiplier=1.0):
        assert isinstance(nonce_multiplier, float), "nonce_multiplier must be decimal"
        self.url = "%s://%s/%s" % (PROTOCOL, HOST, VERSION)
        self.base_url = "%s://%s/" % (PROTOCOL, HOST)
        self.key = key
        self.secret = secret
        self.nonce_multiplier = nonce_multiplier

    def server(self):
        return u"{0:s}://{1:s}/{2:s}".format(PROTOCOL, HOST, VERSION)

    def url_for(self, path, path_arg=None, parameters=None):

        # build the basic url
        url = "%s/%s" % (self.server(), path)

        # If there is a path_arh, interpolate it into the URL.
        # In this case the path that was provided will need to have string
        # interpolation characters in it, such as PATH_TICKER
        if path_arg:
            url = url % (path_arg)

        # Append any parameters to the URL.
        if parameters:
            url = "%s?%s" % (url, self._build_parameters(parameters))

        return url

    def _nonce(self):
        """Returns a nonce used in authentication.
        Nonce must be an increasing number, if the API key has been used
        earlier or other frameworks that have used higher numbers you might
        need to increase the nonce_multiplier."""
        return str(utils.get_nonce(self.nonce_multiplier))

    def _sign_payload(self, payload):
        j = json.dumps(payload)
        data = base64.standard_b64encode(j.encode('utf8'))

        hmc = hmac.new(self.secret.encode('utf8'), data, hashlib.sha384)
        signature = hmc.hexdigest()
        return {
            "X-BFX-APIKEY": self.key,
            "X-BFX-SIGNATURE": signature,
            "X-BFX-PAYLOAD": data
        }

    def _get(self, url):
        response = requests.get(url, timeout=TIMEOUT)
        if response.status_code == 200:
            return response.json()
        else:
            try:
                content = response.json()
            except JSONDecodeError:
                content = response.text()
            raise BitfinexException(response.status_code, response.reason, content)

    def _post(self, endoint, payload, verify=True):
        url = self.url_for(path=endoint)
        signed_payload = self._sign_payload(payload)
        response = requests.post(url, headers=signed_payload, verify=verify)
        if response.status_code == 200:
            return response.json()
        elif response.status_code == 400:
            return response.json()
        else:
            try:
                content = response.json()
            except JSONDecodeError:
                content = response.text()
            raise BitfinexException(response.status_code, response.reason, content)

    def _build_parameters(self, parameters):
        # sort the keys so we can test easily in Python 3.3 (dicts are not
        # ordered)
        keys = list(parameters.keys())
        keys.sort()

        return '&'.join(["%s=%s" % (k, parameters[k]) for k in keys])

    def place_order(self, amount, price, side, ord_type, symbol='btcusd', exchange='bitfinex'):
        """
        .. _new_order:

        `Bitfinex new order <https://docs.bitfinex.com/v1/reference#rest-auth-new-order>`_

        Submit a new Order

        Parameters
        ----------
        amount : float
            Order size: how much you want to buy or sell

        price : float
            Price to buy or sell at. Must be positive. Use random number for market orders.

        side : string
            Either “buy” or “sell”.

        ord_type : string
            Either “market” / “limit” / “stop” / “trailing-stop” / “fill-or-kill” /
            “exchange market” / “exchange limit” / “exchange stop” / “exchange trailing-stop” /
            “exchange fill-or-kill”. (type starting by “exchange ” are exchange orders, others are
            margin trading orders)

        symbol : str
            The `symbol <restv1.html#symbols>`_ you want information about.

        exchange : str
            'bitfinex'

        Returns
        -------
        dict
             ::

                # response
                {
                  "id":448364249,
                  "symbol":"btcusd",
                  "exchange":"bitfinex",
                  "price":"0.01",
                  "avg_execution_price":"0.0",
                  "side":"buy",
                  "type":"exchange limit",
                  "timestamp":"1444272165.252370982",
                  "is_live":true,
                  "is_cancelled":false,
                  "is_hidden":false,
                  "was_forced":false,
                  "original_amount":"0.01",
                  "remaining_amount":"0.01",
                  "executed_amount":"0.0",
                  "order_id":448364249
                }

        Examples
        --------
         ::

            bfx_client.place_order(0.01, 0.01, "buy", "exchange limit", "btcusd")

        """

        payload = {

            "request": "/v1/order/new",
            "nonce": self._nonce(),
            "symbol": symbol,
            "amount": amount,
            "price": price,
            "exchange": exchange,
            "side": side,
            "type": ord_type

        }

        response = self._post("/order/new", payload=payload, verify=True)
        return response

    def place_multiple_orders(self, orders):
        """
        Parameters
        ----------
        orders : list
            Each item in the list is a dict that must have the following items : symbol, amount,
            price, side, type, exchange



        Returns
        -------
        dict
             ::

                // response
                {
                  "order_ids":[
                    {
                        "id":448383727,
                        "symbol":"btcusd",
                        "exchange":"bitfinex",
                        "price":"0.01",
                        "avg_execution_price":"0.0",
                        "side":"buy",
                        "type":"exchange limit",
                        "timestamp":"1444274013.621701916",
                        "is_live":true,
                        "is_cancelled":false,
                        "is_hidden":false,
                        "was_forced":false,
                        "original_amount":"0.01",
                        "remaining_amount":"0.01",
                        "executed_amount":"0.0"
                    },{
                        "id":448383729,
                        "symbol":"btcusd",
                        "exchange":"bitfinex",
                        "price":"0.03",
                        "avg_execution_price":"0.0",
                        "side":"buy",
                        "type":"exchange limit",
                        "timestamp":"1444274013.661297306",
                        "is_live":true,
                        "is_cancelled":false,
                        "is_hidden":false,
                        "was_forced":false,
                        "original_amount":"0.02",
                        "remaining_amount":"0.02",
                        "executed_amount":"0.0"
                    }],
                  "status":"success"
                }

        Examples
        --------
         ::

            # Make a list with 3 orders to buy 100 iota at 3 dollars,100 iota at 4 dollars and
            # 100 iota at 5 dollars
            # The list is sent to the method place_multiple_orders
            orders = []
            for price in range(3, 6):
                print(price)
                payload = {
                    "symbol": 'IOTUSD',
                    "amount": '100',
                    "price": str(price),
                    "exchange": 'bitfinex',
                    "side": 'buy',
                    "type": 'limit'
                }
                orders.append(payload)
            response = bfx_client.place_multiple_orders(orders)
            print(response)
        """

        payload = {
            "request": "/v1/order/new/multi",
            "nonce": self._nonce(),
            "orders": orders
        }
        response = self._post("/order/new/multi", payload=payload, verify=True)
        return response

    def delete_order(self, order_id):
        """`Bitfinex cancel order reference
        <https://docs.bitfinex.com/v1/reference#rest-auth-cancel-order>`_

        Cancel an order.

        Parameters
        ----------
        order_id : int
            The order ID given by `new_order`_ function

        Returns
        -------
        dict
             ::

                {
                  "id":446915287,
                  "symbol":"btcusd",
                  "exchange":null,
                  "price":"239.0",
                  "avg_execution_price":"0.0",
                  "side":"sell",
                  "type":"trailing stop",
                  "timestamp":"1444141982.0",
                  "is_live":true,
                  "is_cancelled":false,
                  "is_hidden":false,
                  "was_forced":false,
                  "original_amount":"1.0",
                  "remaining_amount":"1.0",
                  "executed_amount":"0.0"
                }

        Example
        -------
         ::

            bfx_client.delete_order(448411153)

        """

        payload = {
            "request": "/v1/order/cancel",
            "nonce": self._nonce(),
            "order_id": order_id
        }

        response = self._post("/order/cancel", payload=payload, verify=True)
        return response

    def delete_all_orders(self):
        """`Bitfinex cancel all orders reference
        <https://docs.bitfinex.com/v1/reference#rest-auth-cancel-multiple-orders>`_

        Cancel all orders at once.

        Returns
        -------
        dict
             ::

                {"result":"Orders cancelled"}

        Example
        -------
         ::

            bfx_client.delete_all_orders()

        """
        payload = {
            "request": "/v1/order/cancel/all",
            "nonce": self._nonce(),
        }

        response = self._post("/order/cancel/all", payload=payload, verify=True)
        return response

    def status_order(self, order_id):
        """`Bitfinex status order reference
        <https://docs.bitfinex.com/v1/reference#rest-auth-order-status>`_

        Get the status of an order. Is it active? Was it cancelled?
        To what extent has it been executed? etc.

        Parameters
        ----------
        order_id : int
            The order ID given by `new_order`_ function

        Returns
        -------
        dict
             ::

                {
                  "id":448411153,
                  "symbol":"btcusd",
                  "exchange":null,
                  "price":"0.01",
                  "avg_execution_price":"0.0",
                  "side":"buy",
                  "type":"exchange limit",
                  "timestamp":"1444276570.0",
                  "is_live":false,
                  "is_cancelled":true,
                  "is_hidden":false,
                  "oco_order":null,
                  "was_forced":false,
                  "original_amount":"0.01",
                  "remaining_amount":"0.01",
                  "executed_amount":"0.0"
                }

        Example
        -------
         ::

            bfx_client.status_order(448411153)

        """

        payload = {
            "request": "/v1/order/status",
            "nonce": self._nonce(),
            "order_id": order_id
        }

        response = self._post("/order/status", payload=payload, verify=True)
        return response

    def active_orders(self):
        """`Bitfinex active orders reference
        <https://docs.bitfinex.com/v1/reference#rest-auth-active-orders>`_

        View your active orders.

        Returns
        -------
        list
             ::

                [{
                  "id":448411365,
                  "symbol":"btcusd",
                  "exchange":"bitfinex",
                  "price":"0.02",
                  "avg_execution_price":"0.0",
                  "side":"buy",
                  "type":"exchange limit",
                  "timestamp":"1444276597.0",
                  "is_live":true,
                  "is_cancelled":false,
                  "is_hidden":false,
                  "was_forced":false,
                  "original_amount":"0.02",
                  "remaining_amount":"0.02",
                  "executed_amount":"0.0"
                }]

        Example
        -------
         ::

            bfx_client.active_orders(448411153)

        """
        payload = {
            "request": "/v1/orders",
            "nonce": self._nonce()
        }
        response = self._post("orders", payload=payload, verify=True)
        return response

    def active_positions(self):
        """`Bitfinex active positions reference
        <https://docs.bitfinex.com/v1/reference#rest-auth-active-positions>`_

        View your active positions.

        Returns
        -------
        list
             ::

                [{
                  "id":943715,
                  "symbol":"btcusd",
                  "status":"ACTIVE",
                  "base":"246.94",
                  "amount":"1.0",
                  "timestamp":"1444141857.0",
                  "swap":"0.0",
                  "pl":"-2.22042"
                }]

        Example
        -------
         ::

            bfx_client.active_positions(448411153)

        """
        payload = {
            "request": "/v1/positions",
            "nonce": self._nonce()
        }
        response = self._post("positions", payload=payload, verify=True)
        return response

    def claim_position(self, position_id):
        """`Bitfinex claim position reference
        <https://docs.bitfinex.com/v1/reference#rest-auth-claim-position>`_

        A position can be claimed if: It is a long position: The amount in the last unit of the
        position pair that you have in your trading wallet AND/OR the realized profit of the
        position is greater or equal to the purchase amount of the position
        (base price * position amount) and the funds which need to be returned. For example, for a
        long BTCUSD position, you can claim the position if the amount of USD you have in the
        trading wallet is greater than the base price * the position amount and the funds used.
        It is a short position: The amount in the first unit of the position pair that you have in
        your trading wallet is greater or equal to the amount of the position and the margin
        funding used.

        Parameters
        ----------
        position_id : int
            The position ID

        Returns
        -------
        dict
             ::

                {
                  "id":943715,
                  "symbol":"btcusd",
                  "status":"ACTIVE",
                  "base":"246.94",
                  "amount":"1.0",
                  "timestamp":"1444141857.0",
                  "swap":"0.0",
                  "pl":"-2.2304"
                }

        Example
        -------
         ::

            bfx_client.claim_position(18411153)

        """
        payload = {
            "request": "/v1/position/claim",
            "nonce": self._nonce(),
            "position_id": position_id
        }
        response = self._post("position/claim", payload=payload, verify=True)
        return response

    def close_position(self, position_id):
        """`Bitfinex close position reference
        <https://docs.bitfinex.com/v1/reference#close-position>`_

        Closes the selected position with a market order.

        Parameters
        ----------
        position_id : int
            The position ID

        Returns
        -------
        dict
             ::

                {
                  "message": "",
                    "order": {},
                  "position": {}
                }

        Example
        -------
         ::

            bfx_client.close_position(18411153)

        """
        payload = {
            "request": "/v1/position/close",
            "nonce": self._nonce(),
            "position_id": position_id
        }
        response = self._post("position/close", payload=payload, verify=True)
        return response

    def past_trades(self, timestamp='0.0', symbol='btcusd'):
        """`Bitfinex past trades reference
        <https://docs.bitfinex.com/v1/reference#rest-auth-past-trades>`_

        View your past trades.

        Parameters
        ----------
        timestamp : str
            Trades made before this timestamp won’t be returned.

        symbol : str
            The `symbol <restv1.html#symbols>`_ you want information about.

        Returns
        -------
        list
             ::

                [{
                  "price":"246.94",
                  "amount":"1.0",
                  "timestamp":"1444141857.0",
                  "exchange":"",
                  "type":"Buy",
                  "fee_currency":"USD",
                  "fee_amount":"-0.49388",
                  "tid":11970839,
                  "order_id":446913929
                }]

        Example
        -------
         ::

            bfx_client.past_trades('0.0','btcusd')
            bfx_client.past_trades()

        """

        payload = {
            "request": "/v1/mytrades",
            "nonce": self._nonce(),
            "symbol": symbol,
            "timestamp": timestamp
        }

        response = self._post("/mytrades", payload=payload, verify=True)
        return response

    def place_offer(self, currency, amount, rate, period, direction):
        """
        .. _new_offer:

        `Bitfinex place offer reference
        <https://docs.bitfinex.com/v1/reference#rest-auth-new-offer>`_

        Submit a new Offer

        Parameters
        ----------
        currency : string
            The name of the currency.

        amount : float
            Order size: how much to lend or borrow.

        rate : float
            Rate to lend or borrow at. In percentage per 365 days. (Set to 0 for FRR).

        period : int
            Number of days of the funding contract (in days)

        direction : string
            Either “lend” or “loan”.

        Returns
        -------
        dict
             ::

                {
                  "id":13800585,
                  "currency":"USD",
                  "rate":"20.0",
                  "period":2,
                  "direction":"lend",
                  "timestamp":"1444279698.21175971",
                  "is_live":true,
                  "is_cancelled":false,
                  "original_amount":"50.0",
                  "remaining_amount":"50.0",
                  "executed_amount":"0.0",
                  "offer_id":13800585
                }

        Example
        -------
         ::

            bfx_client.place_offer("USD",50.0,20.0,2,"lend")

        """
        payload = {
            "request": "/v1/offer/new",
            "nonce": self._nonce(),
            "currency": currency,
            "amount": amount,
            "rate": rate,
            "period": period,
            "direction": direction
        }

        response = self._post("/offer/new", payload=payload, verify=True)
        return response

    def cancel_offer(self, offer_id):
        """`Bitfinex cancel offer reference
        <https://docs.bitfinex.com/v1/reference#rest-auth-cancel-offer>`_

        Cancel an offer.

        Parameters
        ----------
        offer_id : int
            The offer ID given by `new_offer`_ function

        Returns
        -------
        dict
             ::

                {
                  "id":13800585,
                  "currency":"USD",
                  "rate":"20.0",
                  "period":2,
                  "direction":"lend",
                  "timestamp":"1444279698.0",
                  "is_live":true,
                  "is_cancelled":false,
                  "original_amount":"50.0",
                  "remaining_amount":"50.0",
                  "executed_amount":"0.0"
                }

        Example
        -------
         ::

            bfx_client.cancel_offer(1231153)

        """
        payload = {
            "request": "/v1/offer/cancel",
            "nonce": self._nonce(),
            "offer_id": offer_id
        }

        response = self._post("/offer/cancel", payload=payload, verify=True)
        return response

    def status_offer(self, offer_id):
        """`Bitfinex status offer reference
        <https://docs.bitfinex.com/v1/reference#rest-auth-offer-status>`_

        Get the status of an offer. Is it active? Was it cancelled?
        To what extent has it been executed? etc.

        Parameters
        ----------
        offer_id : int
            The offer ID given by `new_offer`_ function

        Returns
        -------
        dict
             ::

                {
                  "id":13800585,
                  "currency":"USD",
                  "rate":"20.0",
                  "period":2,
                  "direction":"lend",
                  "timestamp":"1444279698.0",
                  "is_live":false,
                  "is_cancelled":true,
                  "original_amount":"50.0",
                  "remaining_amount":"50.0",
                  "executed_amount":"0.0"
                }

        Example
        -------
         ::

            bfx_client.status_offer(354352)

        """
        payload = {
            "request": "/v1/offer/status",
            "nonce": self._nonce(),
            "offer_id": offer_id
        }

        response = self._post("/offer/status", payload=payload, verify=True)
        return response

    def offers_history(self):
        """`Bitfinex offers history reference
        <https://docs.bitfinex.com/v1/reference#rest-auth-offers-hist>`_

        View your latest inactive offers. Limited to last 3 days and 1 request per minute.

        Returns
        -------
        list
             ::

                [{
                  "id":13800719,
                  "currency":"USD",
                  "rate":"31.39",
                  "period":2,
                  "direction":"lend",
                  "timestamp":"1444280237.0",
                  "is_live":false,
                  "is_cancelled":true,
                  "original_amount":"50.0",
                  "remaining_amount":"50.0",
                  "executed_amount":"0.0"
                }]

        Example
        -------
         ::

            bfx_client.offers_history()

        """

        payload = {
            "request": "/v1/offers/hist",
            "nonce": self._nonce()
        }
        response = self._post("/offers/hist", payload=payload, verify=True)
        return response

    def active_offers(self):
        """`Bitfinex active offers reference
        <https://docs.bitfinex.com/v1/reference#rest-auth-offers>`_

        View your active offers.

        Returns
        -------
        list
             ::

                [{
                  "id":13800719,
                  "currency":"USD",
                  "rate":"31.39",
                  "period":2,
                  "direction":"lend",
                  "timestamp":"1444280237.0",
                  "is_live":true,
                  "is_cancelled":false,
                  "original_amount":"50.0",
                  "remaining_amount":"50.0",
                  "executed_amount":"0.0"
                }]

        Example
        -------
         ::

            bfx_client.active_offers()

        """
        payload = {
            "request": "/v1/offers",
            "nonce": self._nonce()
        }

        response = self._post("/offers", payload=payload, verify=True)
        return response

    def balances(self):
        """`Bitfinex balances reference
        <https://docs.bitfinex.com/v1/reference#rest-auth-wallet-balances>`_

        See your balances

        Returns
        -------
        list
             ::

                [
                    {"type":"deposit", "currency":"btc", "amount":"0.0", "available":"0.0"},
                    {"type":"deposit", "currency":"usd", "amount":"1.0", "available":"1.0"},
                    {"type":"exchange", "currency":"btc", "amount":"1", "available":"1"},
                    {"type":"exchange", "currency":"usd", "amount":"1", "available":"1"},
                    {"type":"trading", "currency":"btc", "amount":"1", "available":"1"},
                    {"type":"trading", "currency":"usd", "amount":"1", "available":"1"},
                ...
                ]

        Example
        -------
         ::

            bfx_client.balances()

        """
        payload = {
            "request": "/v1/balances",
            "nonce": self._nonce()
        }
        response = self._post("/balances", payload=payload, verify=True)
        return response

    def history(self, currency, since=0, until=9999999999, limit=500, wallet='exchange'):
        """`Bitfinex balance history reference
        <https://docs.bitfinex.com/v1/reference#rest-auth-wallet-balances>`_

        View all of your balance ledger entries.


        Parameters
        ----------
        currency : str
            The currency to look for.

        since : Optional str
            Return only the history after this timestamp. (Max 3 months before until)

        until : Optional str
            Return only the history before this timestamp. (Default now)

        limit : Optional int
            Limit the number of entries to return.

        wallet : Optional str
            Return only entries that took place in this wallet.
            Accepted inputs are: “trading”, “exchange”, “deposit”.

        Returns
        -------
        list
             ::

                [{
                  "currency":"USD",
                  "amount":"-246.94",
                  "balance":"515.4476526",
                  "description":"Position claimed @ 245.2 on wallet trading",
                  "timestamp":"1444277602.0"
                }]

        Example
        -------
         ::

            bfx_client.history('USD')

        """
        payload = {
            "request": "/v1/history",
            "nonce": self._nonce(),
            "currency": currency,
            "since": since,
            "until": until,
            "limit": limit,
            "wallet": wallet
        }
        response = self._post("/history", payload=payload, verify=True)
        return response

    def movements(self, currency, start=0, end=9999999999, limit=10000, method='bitcoin'):
        """`Bitfinex Deposit-Withdrawal History reference
        <https://docs.bitfinex.com/v1/reference#rest-auth-wallet-balances>`_

        View your past deposits/withdrawals.

        Parameters
        ----------
        currency : str
            The currency to look for.

        start : Optional str
            Return only the history after this timestamp. (Max 3 months before until)

        end : Optional str
            Return only the history before this timestamp. (Default now)

        limit : Optional int
            Limit the number of entries to return.

        method : str
            The method of the deposit/withdrawal (can be "bitcoin", "litecoin", "iota", "wire").

        Returns
        -------
        list
             ::

                [{
                  "id":581183,
                  "txid": 123456,
                  "currency":"BTC",
                  "method":"BITCOIN",
                  "type":"WITHDRAWAL",
                  "amount":".01",
                  "description":"3QXYWgRGX2BPYBpUDBssGbeWEa5zq6snBZ, offchain transfer ",
                  "address":"3QXYWgRGX2BPYBpUDBssGbeWEa5zq6snBZ",
                  "status":"COMPLETED",
                  "timestamp":"1443833327.0",
                  "timestamp_created": "1443833327.1",
                  "fee": 0.1
                }]

        Example
        -------
         ::

            bfx_client.movements('USD')

        """
        payload = {
            "request": "/v1/history/movements",
            "nonce": self._nonce(),
            "currency": currency,
            "since": start,
            "until": end,
            "limit": limit,
            "method": method
        }
        response = self._post("/history/movements", payload=payload, verify=True)
        return response

    def symbols(self):
        """
        .. _symbols:

        `Bitfinex symbols reference
        <https://docs.bitfinex.com/v1/reference#rest-public-symbols>`_

        A list of symbol names.

        Returns
        -------
        list
             ::

                [
                  "btcusd",
                  "ltcusd",
                  "ltcbtc",
                  ...
                ]

        Example
        -------
         ::

            bfx_client.symbols()

        """
        return self._get(self.url_for(PATH_SYMBOLS))

    def symbols_details(self):
        """`Bitfinex symbols details reference
        <https://docs.bitfinex.com/v1/reference#rest-public-symbol-details>`_

        Get a list of valid symbol IDs and the pair details.

        Returns
        -------
        list
             ::

                [{
                  "pair":"btcusd",
                  "price_precision":5,
                  "initial_margin":"30.0",
                  "minimum_margin":"15.0",
                  "maximum_order_size":"2000.0",
                  "minimum_order_size":"0.01",
                  "expiration":"NA"
                },{
                  "pair":"ltcusd",
                  "price_precision":5,
                  "initial_margin":"30.0",
                  "minimum_margin":"15.0",
                  "maximum_order_size":"5000.0",
                  "minimum_order_size":"0.1",
                  "expiration":"NA"
                },{
                  "pair":"ltcbtc",
                  "price_precision":5,
                  "initial_margin":"30.0",
                  "minimum_margin":"15.0",
                  "maximum_order_size":"5000.0",
                  "minimum_order_size":"0.1",
                  "expiration":"NA"
                },
                ...
                ]

        Example
        -------
         ::

            bfx_client.symbols_details()

        """
        return self._get(self.url_for("symbols_details"))

    def ticker(self, symbol):
        """`Bitfinex ticker reference
        <https://docs.bitfinex.com/v1/reference#rest-public-ticker>`_

        The ticker is a high level overview of the state of the market. It shows you the current
        best bid and ask, as well as the last trade price. It also includes information such as
        daily volume and how much the price has moved over the last day.

        Parameters
        ----------
        symbol : str
            The `symbol <restv1.html#symbols>`_ you want information about.

        Returns
        -------
        dict
             ::

                {
                  "mid":"244.755",
                  "bid":"244.75",
                  "ask":"244.76",
                  "last_price":"244.82",
                  "low":"244.2",
                  "high":"248.19",
                  "volume":"7842.11542563",
                  "timestamp":"1444253422.348340958"
                }

        Example
        -------
         ::

            bfx_client.ticker("BTCUSD")


        """

        return self._get(self.url_for(PATH_TICKER, (symbol)))

    def today(self, symbol):
        """.. _today:

        Get today stats

        Parameters
        ----------
        symbol : str
            The `symbol <restv1.html#symbols>`_ you want information about.

        Returns
        -------
        dict
             ::

                 {"low":"550.09","high":"572.2398","volume":"7305.33119836"}

        Example
        -------
         ::

            bfx_client.today("BTCUSD")

        """

        return self._get(self.url_for(PATH_TODAY, (symbol)))

    def stats(self, symbol):
        """`Bitfinex stats reference
        <https://docs.bitfinex.com/v1/reference#rest-public-stats>`_

        Various statistics about the requested pair.

        Parameters
        ----------
        symbol : str
            The `symbol <restv1.html#symbols>`_ you want information about.

        Returns
        -------
        dict
             ::

                [{
                  "period":1,
                  "volume":"7967.96766158"
                },{
                  "period":7,
                  "volume":"55938.67260266"
                },{
                  "period":30,
                  "volume":"275148.09653645"
                }]

        Example
        -------
         ::

            bfx_client.stats("BTCUSD")

        """
        data = self._get(self.url_for(PATH_STATS, (symbol)))

        for period in data:

            for key, value in period.items():
                if key == 'period':
                    new_value = int(value)
                elif key == 'volume':
                    new_value = value

                period[key] = new_value

        return data

    def lendbook(self, currency, parameters=None):
        """`Bitfinex Fundingbook reference
        <https://docs.bitfinex.com/v1/reference#rest-public-fundingbook>`_

        Get the full margin funding book

        Parameters
        ----------
        currency : str
            Currency

        Returns
        -------
        dict
             ::

                {
                  "bids":[{
                      'rate': '7.0482',
                      'amount': '244575.00836875',
                      'period': 30,
                      'timestamp': '1539157649.0',
                      'frr': False
                  }]
                  "asks":[{
                      'rate': '5.6831',
                      'amount': '63.5744',
                      'period': 2,
                      'timestamp':
                      '1539165059.0',
                      'frr': False
                  }]
                }

        Example
        -------
         ::

            bfx_client.lendbook("USD")

        """
        data = self._get(self.url_for(PATH_LENDBOOK, path_arg=currency, parameters=parameters))

        for lend_type in data.keys():

            for lend in data[lend_type]:

                for key, value in lend.items():
                    if key in ['rate', 'amount', 'timestamp']:
                        new_value = value
                    elif key == 'period':
                        new_value = int(value)
                    elif key == 'frr':
                        new_value = value == 'Yes'

                    lend[key] = new_value

        return data

    def order_book(self, symbol, parameters=None):
        """`Bitfinex Orderbook reference
        <https://docs.bitfinex.com/v1/reference#rest-public-orderbook>`_

        Get the full order book.

        Parameters
        ----------
        currency : str
            Currency

        Returns
        -------
        dict
             ::

                {
                  "bids":[{
                    "price":"574.61",
                    "amount":"0.1439327",
                    "timestamp":"1472506127.0"
                  }],
                  "asks":[{
                    "price":"574.62",
                    "amount":"19.1334",
                    "timestamp":"1472506126.0"
                  }]
                }

        Example
        -------
         ::

            bfx_client.order_book("BTCUSD")

        """
        data = self._get(self.url_for(PATH_ORDERBOOK, path_arg=symbol, parameters=parameters))

        for type_ in data.keys():
            for list_ in data[type_]:
                for key, value in list_.items():
                    list_[key] = value

        return data

    def key_permissions(self):
        """`Bitfinex API Key Permissions
        <https://docs.bitfinex.com/reference#auth-key-permissions>`_

        Check the permissions of the key being used to generate this request.

        Returns
        -------
        dict
             ::

                {
                    'account': {'read': True, 'write': False},
                    'history': {'read': True, 'write': False},
                    'orders': {'read': True, 'write': False},
                    'positions': {'read': True, 'write': False},
                    'funding': {'read': True, 'write': False},
                    'wallets': {'read': True, 'write': False},
                    'withdraw': {'read': False, 'write': False}
                }

        Example
        -------
         ::

            bfx_client.order_book("BTCUSD")

        """
        data = self._post(PATH_KEYINFO, {'request': '/v1/key_info', 'nonce': self._nonce()})

        return data


class TradeClient(Client):
    """Added for backward compatibility"""
    pass



#!/usr/bin/python
"""Bitfinex Rest API V2 implementation"""
# pylint: disable=R0904

from __future__ import absolute_import
import json
from json.decoder import JSONDecodeError
import hmac
import hashlib
import requests
from bitfinex import utils

PROTOCOL = "https"
HOST = "api.bitfinex.com"
VERSION = "v2"


# HTTP request timeout in seconds
TIMEOUT = 5.0


class BitfinexException(Exception):
    """
    Exception handler
    """
    pass


class Client:
    """Client for the bitfinex.com API REST V2.
    Link for official bitfinex documentation :

    `Bitfinex rest2 docs <https://bitfinex.readme.io/v2/docs>`_

    `Bitfinex rest2 reference <https://bitfinex.readme.io/v2/reference>`_

    Parameters
    ----------
    key : str
        Bitfinex api key

    secret : str
        Bitfinex api secret

    nonce_multiplier : Optional float
        Multiply nonce by this number

    Examples
    --------
     ::

        bfx_client = Client(key,secret)

        bfx_client = Client(key,secret,2.0)
    """

    def __init__(self, key=None, secret=None, nonce_multiplier=1.0):
        """
        Object initialisation takes 2 mandatory arguments key and secret and a optional one
        nonce_multiplier
        """

        assert isinstance(nonce_multiplier, float), "nonce_multiplier must be decimal"
        self.base_url = "%s://%s/" % (PROTOCOL, HOST)
        self.key = key
        self.secret = secret
        self.nonce_multiplier = nonce_multiplier

    def _nonce(self):
        """Returns a nonce used in authentication.
        Nonce must be an increasing number, if the API key has been used
        earlier or other frameworks that have used higher numbers you might
        need to increase the nonce_multiplier."""
        return str(utils.get_nonce(self.nonce_multiplier))

    def _headers(self, path, nonce, body):
        """
        create signed headers
        """
        signature = "/api/{}{}{}".format(path, nonce, body)
        hmc = hmac.new(self.secret.encode('utf8'), signature.encode('utf8'), hashlib.sha384)
        signature = hmc.hexdigest()

        return {
            "bfx-nonce": nonce,
            "bfx-apikey": self.key,
            "bfx-signature": signature,
            "content-type": "application/json"
        }

    def _post(self, path, payload, verify=False):
        """
        Send post request to bitfinex
        """
        nonce = self._nonce()
        headers = self._headers(path, nonce, payload)
        response = requests.post(self.base_url + path, headers=headers, data=payload, verify=verify)

        if response.status_code == 200:
            return response.json()
        else:
            try:
                content = response.json()
            except JSONDecodeError:
                content = response.text()
            raise BitfinexException(response.status_code, response.reason, content)

    def _get(self, path, **params):
        """
        Send get request to bitfinex
        """
        url = self.base_url + path
        response = requests.get(url, timeout=TIMEOUT, params=params)
        if response.status_code == 200:
            return response.json()
        else:
            try:
                content = response.json()
            except JSONDecodeError:
                content = response.text()
            raise BitfinexException(response.status_code, response.reason, content)

    # REST PUBLIC ENDPOINTS
    def platform_status(self):
        """
        .. _platform_status:

        `Bitfinex platform_status reference
        <https://bitfinex.readme.io/v2/reference#rest-public-platform-status>`_

        Get the current status of the platform. Maintenance periods last for just few minutes and
        might be necessary from time to time during upgrades of core components of our
        infrastructure. Even if rare it is important to have a way to notify users. For a real-time
        notification we suggest to use websockets and listen to events 20060/20061

        Returns
        -------
        int
            - 1 = operative
            - 0 = maintenance

        Example
        -------
         ::

            bfx_client.platform_status()

        """
        path = "v2/platform/status"
        response = self._get(path)
        return response

    def tickers(self, symbol_list):
        """`Bitfinex tickers reference
        <https://bitfinex.readme.io/v2/reference#rest-public-tickers>`_

        The ticker is a high level overview of the state of the market. It shows you the current
        best bid and ask, as well as the last trade price.It also includes information such as daily
        volume and how much the price has moved over the last day.

        Parameters
        ----------
        symbol_list : list
            The symbols you want information about as a comma separated list,
            or ALL for every symbol.

        Returns
        -------
        list
             ::

                [
                  # on trading pairs (ex. tBTCUSD)
                  [
                    SYMBOL,
                    BID,
                    BID_SIZE,
                    ASK,
                    ASK_SIZE,
                    DAILY_CHANGE,
                    DAILY_CHANGE_PERC,
                    LAST_PRICE,
                    VOLUME,
                    HIGH,
                    LOW
                  ],
                  # on funding currencies (ex. fUSD)
                  [
                    SYMBOL,
                    FRR,
                    BID,
                    BID_SIZE,
                    BID_PERIOD,
                    ASK,
                    ASK_SIZE,
                    ASK_PERIOD,
                    DAILY_CHANGE,
                    DAILY_CHANGE_PERC,
                    LAST_PRICE,
                    VOLUME,
                    HIGH,
                    LOW
                  ],
                  ...
                ]

        Note
        ----
            ================= ===== ================================================================
            Field             Type  Description
            ================= ===== ================================================================
            FRR               float Flash Return Rate - average of all fixed rate funding over the
                                    last hour
            BID               float Price of last highest bid
            BID_PERIOD        int   Bid period covered in days
            BID_SIZE          float Sum of the 25 highest bid sizes
            ASK               float Price of last lowest ask
            ASK_PERIOD        int   Ask period covered in days
            ASK_SIZE          float Sum of the 25 lowest ask sizes
            DAILY_CHANGE      float Amount that the last price has changed since yesterday
            DAILY_CHANGE_PERC float Amount that the price has changed expressed in percentage terms
            LAST_PRICE        float Price of the last trade
            VOLUME            float Daily volume
            HIGH              float Daily high
            LOW               float Daily low
            ================= ===== ================================================================

        Examples
        --------
         ::

            bfx_client.tickers(['tIOTUSD', 'fIOT'])
            bfx_client.tickers(['tBTCUSD'])
            bfx_client.tickers(['ALL'])

        """
        assert isinstance(symbol_list, list), "symbol_list must be of type list"
        assert symbol_list, "symbol_list must have at least one symbol"
        path = "v2/tickers?symbols={}".format(",".join(symbol_list))
        response = self._get(path)
        return response

    def ticker(self, symbol):
        """`Bitfinex ticker reference
        <https://bitfinex.readme.io/v2/reference#rest-public-ticker>`_

        The ticker is a high level overview of the state of the market.It shows you the current best
        bid and ask, as well as the last trade price.It also includes information such as daily
        volume and how much the price has moved over the last day.

        Parameters
        ----------
        symbol : str
            The symbol you want information about.
            You can find the list of valid symbols by calling the `symbols <restv1.html#symbols>`_
            method

        Returns
        -------
        list
             ::

                # on trading pairs (ex. tBTCUSD)
                [
                  BID,
                  BID_SIZE,
                  ASK,
                  ASK_SIZE,
                  DAILY_CHANGE,
                  DAILY_CHANGE_PERC,
                  LAST_PRICE,
                  VOLUME,
                  HIGH,
                  LOW
                ]
                # on funding currencies (ex. fUSD)
                [
                  FRR,
                  BID,
                  BID_SIZE,
                  BID_PERIOD,
                  ASK,
                  ASK_SIZE,
                  ASK_PERIOD,
                  DAILY_CHANGE,
                  DAILY_CHANGE_PERC,
                  LAST_PRICE,
                  VOLUME,
                  HIGH,
                  LOW
                ]

        Examples
        --------
         ::

            bfx_client.ticker('tIOTUSD')
            bfx_client.ticker('fIOT')
            bfx_client.ticker('tBTCUSD')

        """
        path = "v2/ticker/{}".format(symbol)
        response = self._get(path)
        return response

    def trades(self, symbol):
        """`Bitfinex trades reference
        <https://bitfinex.readme.io/v2/reference#rest-public-trades>`_

        Trades endpoint includes all the pertinent details of the trade, such as price,
        size and time.

        Parameters
        ----------
        symbol : str
            The symbol you want information about.
            You can find the list of valid symbols by calling the `symbols <restv1.html#symbols>`_
            method

        Returns
        -------
        list
             ::

                # on trading pairs (ex. tBTCUSD)
                [
                  [
                    ID,
                    MTS,
                    AMOUNT,
                    PRICE
                  ]
                ]
                # on funding currencies (ex. fUSD)
                [
                  [
                    ID,
                    MTS,
                    AMOUNT,
                    RATE,
                    PERIOD
                  ]
                ]

        Examples
        --------
         ::

            bfx_client.trades('tIOTUSD')
            bfx_client.trades('fIOT')
            bfx_client.trades('tBTCUSD')

        """
        path = "v2/trades/{}/hist".format(symbol)
        response = self._get(path)
        return response

    def books(self, symbol, precision="P0"):
        """`Bitfinex books reference
        <https://bitfinex.readme.io/v2/reference#rest-public-books>`_

        The Order Books channel allow you to keep track of the state of the Bitfinex order book.
        It is provided on a price aggregated basis, with customizable precision.

        Parameters
        ----------
        symbol : str
            The `symbol <restv1.html#symbols>`_ you want information about.

        precision : Optional str
            Level of price aggregation (P0, P1, P2, P3, R0).
            R0 means "gets the raw orderbook".

        Returns
        -------
        list
             ::

                # on trading pairs (ex. tBTCUSD)
                [
                  [
                    PRICE,
                    COUNT,
                    AMOUNT
                  ]
                ]
                # on funding currencies (ex. fUSD)
                [
                  [
                    RATE,
                    PERIOD,
                    COUNT,
                    AMOUNT
                  ]
                ]

        Examples
        --------
         ::

            bfx_client.books('tIOTUSD')
            bfx_client.books('fIOT')
            bfx_client.books('tBTCUSD')

        """
        path = "v2/book/{}/{}".format(symbol, precision)
        response = self._get(path)
        return response

    def stats(self, **kwargs):
        """`Bitfinex stats reference
        <https://bitfinex.readme.io/v2/reference#rest-public-stats>`_

        Various statistics about the requested pair.

        Parameters
        ----------
        Key : str
            Allowed values: "funding.size", "credits.size", "credits.size.sym",
            "pos.size"

        Size : str
            Available values: '1m'

        Symbol : str
            The symbol you want information about.

        Symbol2 : str
            The symbol you want information about.

        Side : str
            Available values: "long", "short"

        Section : str
            Available values: "last", "hist"

        sort : str
            if = 1 it sorts results returned with old > new

        Returns
        -------
        list
             ::

                # response with Section = "last"
                [
                  MTS,
                  VALUE
                ]
                # response with Section = "hist"
                [
                  [ MTS, VALUE ],
                  ...
                ]

        Examples
        --------
         ::

            PARAMS = {
                'key': 'funding.size',
                'size': '1m',
                'symbol': 'fUSD',
                'section': 'hist',
                'sort': '0'
            }
            bfx_client.stats(**PARAMS)              # statistics

            PARAMS = {
                'key': 'credits.size',
                'size': '1m',
                'symbol': 'fUSD',
                'section': 'hist',
                'sort': '0'
            }
            bfx_client.stats(**PARAMS)              # statistics

            PARAMS = {
                'key': 'pos.size',
                'size': '1m',
                'symbol': 'tIOTUSD',
                'side': 'short',
                'section': 'hist',
                'sort': '0'
            }
            bfx_client.stats(**PARAMS)              # statistics

            PARAMS = {
                'key': 'credits.size.sym',
                'size': '1m',
                'symbol': 'fUSD',
                'symbol2': 'tBTCUSD',
                'section': 'hist',
                'sort': '0'
            }

        """
        key_values = ['funding.size', 'credits.size', 'credits.size.sym', 'pos.size']
        if kwargs['key'] not in key_values:
            key_values = " ".join(key_values)
            msg = "Key must have one of the following values : {}".format(key_values)
            raise ValueError(msg)

        common_stats_url = "v2/stats1/{}:{}:{}".format(
            kwargs['key'],
            kwargs['size'],
            kwargs['symbol']
        )

        if kwargs['key'] == 'pos.size':
            custom_stats_url = ":{}/{}?sort={}".format(
                kwargs['side'],
                kwargs['section'],
                str(kwargs['sort'])
            )

        if kwargs['key'] in ['funding.size', 'credits.size']:
            custom_stats_url = "/{}?sort={}".format(
                kwargs['section'],
                str(kwargs['sort'])
            )

        if kwargs['key'] == 'credits.size.sym':
            custom_stats_url = ":{}/{}?sort={}".format(
                kwargs['symbol2'],
                kwargs['section'],
                str(kwargs['sort'])
            )

        path = "".join([common_stats_url, custom_stats_url])

        response = self._get(path)
        return response

    def candles(self, *args, **kwargs):
        """`Bitfinex candles reference
        <https://bitfinex.readme.io/v2/reference#rest-public-candles>`_

        Provides a way to access charting candle info

        Parameters
        ----------
        timeFrame : str
            Available values: '1m', '5m', '15m', '30m', '1h', '3h', '6h', '12h',
            '1D', '7D', '14D', '1M'

        symbol : str
            The symbol you want information about.

        section : str
            Available values: "last", "hist"

        limit : int
            Number of candles requested

        start : str
            Filter start (ms)

        end : str
            Filter end (ms)

        sort : int
            if = 1 it sorts results returned with old > new

        Returns
        -------
        list
             ::

                # response with Section = "last"
                [
                  MTS,
                  OPEN,
                  CLOSE,
                  HIGH,
                  LOW,
                  VOLUME
                ]

                # response with Section = "hist"
                [
                  [ MTS, OPEN, CLOSE, HIGH, LOW, VOLUME ],
                  ...
                ]

        Examples
        --------
         ::

            # 1 minute candles
            bfx_client.candles("1m", "tBTCUSD", "hist")

            # 1 hour candles , limit to 1 candle
            bfx_client.candles("1h", "tBTCUSD", "hist", limit='1')
            bfx_client.candles("1h", "tBTCUSD", "last")

        """
        margs = list(args)
        section = margs.pop()
        path = "v2/candles/trade"
        for arg in margs:
            path = path + ":" + arg
        path += "/{}".format(section)
        response = self._get(path, **kwargs)
        return response

    # REST CALCULATION ENDPOINTS
    def market_average_price(self, **kwargs):
        """`Bitfinex market average price reference
        <https://bitfinex.readme.io/v2/reference#rest-calc-market-average-price>`_

        Calculate the average execution rate for Trading or Margin funding.

        Parameters
        ----------
        symbol : str
            The `symbol <restv1.html#symbols>`_ you want information about.

        amount : str
            Amount. Positive for buy, negative for sell (ex. "1.123")

        period : Optional int
            Maximum period for Margin Funding

        rate_limit : str
            Limit rate/price (ex. "1000.5")

        Returns
        -------
        list
             ::

                [RATE_AVG, AMOUNT]

        Example
        -------
         ::

            bfx_client.market_average_price(symbol="tBTCUSD", amount="100", period="1m")

        """
        body = kwargs
        raw_body = json.dumps(body)
        path = "v2/calc/trade/avg"
        response = self._post(path, raw_body, verify=True)
        return response

    def foreign_exchange_rate(self, **kwargs):
        """`Bitfinex foreign exchange rate reference
        <https://bitfinex.readme.io/v2/reference#foreign-exchange-rate>`_


        Parameters
        ----------
        ccy1 : str
            First currency

        ccy2 : str
            Second currency

        Returns
        -------
        list
             ::

                [ CURRENT_RATE ]

        Example
        -------
         ::

            bfx_client.foreign_exchange_rate(ccy1="IOT", ccy2="USD")

        """
        body = kwargs
        raw_body = json.dumps(body)
        path = "v2/calc/fx"
        response = self._post(path, raw_body, verify=True)
        return response

    # REST AUTHENTICATED ENDPOINTS
    def wallets_balance(self):
        """`Bitfinex wallets balance reference
        <https://bitfinex.readme.io/v2/reference#rest-auth-wallets>`_

        Get account wallets

        Returns
        -------
        list
             ::

                [
                  [
                    WALLET_TYPE,
                    CURRENCY,
                    BALANCE,
                    UNSETTLED_INTEREST,
                    BALANCE_AVAILABLE,
                    ...
                  ],
                  ...
                ]

        Example
        -------
         ::

            bfx_client.wallets_balance()

        """

        body = {}
        raw_body = json.dumps(body)
        path = "v2/auth/r/wallets"
        response = self._post(path, raw_body, verify=True)
        return response

    def active_orders(self, trade_pair=""):
        """`Bitfinex active orders reference
        <https://bitfinex.readme.io/v2/reference#rest-auth-orders>`_

        Fetch active orders using rest api v2

        Parameters
        ----------
        symbol : Optional str
            The `symbol <restv1.html#symbols>`_ you want information about.

        Returns
        -------
        list
             ::

                [
                  [
                    ID,
                    GID,
                    CID,
                    SYMBOL,
                    MTS_CREATE,
                    MTS_UPDATE,
                    AMOUNT,
                    AMOUNT_ORIG,
                    TYPE,
                    TYPE_PREV,
                    _PLACEHOLDER,
                    _PLACEHOLDER,
                    FLAGS,
                    STATUS,
                    _PLACEHOLDER,
                    _PLACEHOLDER,
                    PRICE,
                    PRICE_AVG,
                    PRICE_TRAILING,
                    PRICE_AUX_LIMIT,
                    _PLACEHOLDER,
                    _PLACEHOLDER,
                    _PLACEHOLDER,
                    HIDDEN,
                    PLACED_ID,
                    ...
                  ],
                  ...
                ]

        Examples
        --------
         ::

            bfx_client.active_orders("tIOTUSD")
            bfx_client.active_orders()

        """

        body = {}
        raw_body = json.dumps(body)
        path = "v2/auth/r/orders/{}".format(trade_pair)
        response = self._post(path, raw_body, verify=True)
        return response

    def orders_history(self, trade_pair, **kwargs):
        """`Bitfinex orders history reference
        <https://bitfinex.readme.io/v2/reference#orders-history>`_

        Returns the most recent closed or canceled orders up to circa two weeks ago

        Parameters
        ----------
        symbol : str
            The `symbol <restv1.html#symbols>`_ you want information about.

        start : Optional int
            Millisecond start time

        end : Optional int
            Millisecond end time

        limit : Optional int
            Number of records

        sort : Optional int
            set to 1 to get results in ascending order or -1 for descending

        Returns
        -------
        list
             ::

                [
                  [
                    ID,
                    GID,
                    CID,
                    SYMBOL,
                    MTS_CREATE,
                    MTS_UPDATE,
                    AMOUNT,
                    AMOUNT_ORIG,
                    TYPE,
                    TYPE_PREV,
                    _PLACEHOLDER,
                    _PLACEHOLDER,
                    FLAGS,
                    STATUS,
                    _PLACEHOLDER,
                    _PLACEHOLDER,
                    PRICE,
                    PRICE_AVG,
                    PRICE_TRAILING,
                    PRICE_AUX_LIMIT,
                    _PLACEHOLDER,
                    _PLACEHOLDER,
                    _PLACEHOLDER,
                    NOTIFY,
                    HIDDEN,
                    PLACED_ID,
                    ...
                  ],
                  ...
                ]

        Example
        -------
         ::

            bfx_client.orders_history("tIOTUSD")

        """

        body = kwargs
        raw_body = json.dumps(body)
        path = "v2/auth/r/orders/{}/hist".format(trade_pair)
        response = self._post(path, raw_body, verify=True)
        return response

    def order_trades(self, trade_pair, order_id):
        """`Bitfinex order trades reference
        <https://bitfinex.readme.io/v2/reference#rest-auth-order-trades>`_

        Get Trades generated by an Order

        Parameters
        ----------
        symbol : str
            The `symbol <restv1.html#symbols>`_ you want information about.

        orderid : int
            Order id

        Returns
        -------
        list
             ::

                [
                  [
                    ID,
                    PAIR,
                    MTS_CREATE,
                    ORDER_ID,
                    EXEC_AMOUNT,
                    EXEC_PRICE,
                    _PLACEHOLDER,
                    _PLACEHOLDER,
                    MAKER,
                    FEE,
                    FEE_CURRENCY,
                    ...
                  ],
                  ...
                ]

        Example
        -------
         ::

            bfx_client.order_trades("tIOTUSD", 14395751815)

        """
        body = {}
        raw_body = json.dumps(body)
        path = "v2/auth/r/order/{}:{}/trades".format(trade_pair, order_id)
        response = self._post(path, raw_body, verify=True)
        return response

    def trades_history(self, trade_pair=None, **kwargs):
        """`Bitfinex trades history reference
        <https://docs.bitfinex.com/v2/reference#rest-auth-trades-hist>`_

        List of trades

        Parameters
        ----------
        trade_pair : Optional str
            The `symbol <restv1.html#symbols>`_ you want information about.
            Omit for all symbols

        start : Optional int
            Millisecond start time

        end : Optional int
            Millisecond end time

        limit : Optional int
            Number of records

        Returns
        -------
        list
             ::

                [
                  [
                    ID,
                    PAIR,
                    MTS_CREATE,
                    ORDER_ID,
                    EXEC_AMOUNT,
                    EXEC_PRICE,
                    ORDER_TYPE,
                    ORDER_PRICE,
                    MAKER,
                    FEE,
                    FEE_CURRENCY,
                    ...
                  ],
                  ...
                ]

        Examples
        --------
         ::

            bfx_client.trades_history('tIOTUSD', limit=10)

            TH = BTFXCLIENT.trades_history("tIOTUSD")
            for trade in TH:
                print(trade)

        """

        body = kwargs
        raw_body = json.dumps(body)
        if trade_pair is None:
            path = "v2/auth/r/trades/hist"  # will load history for all pairs
        else:
            path = "v2/auth/r/trades/{}/hist".format(trade_pair)
        response = self._post(path, raw_body, verify=True)
        return response

    def active_positions(self):
        """`Bitfinex positions reference
        <https://bitfinex.readme.io/v2/reference#rest-auth-positions>`_

        Get active positions

        Returns
        -------
        list
             ::

                [
                  [
                    SYMBOL,
                    STATUS,
                    AMOUNT,
                    BASE_PRICE,
                    MARGIN_FUNDING,
                    MARGIN_FUNDING_TYPE,
                    PL,
                    PL_PERC,
                    PRICE_LIQ,
                    LEVERAGE
                    ...
                  ],
                  ...
                ]

        Example
        -------
         ::

            bfx_client.active_positions()

        """
        body = {}
        raw_body = json.dumps(body)
        path = "v2/auth/r/positions"
        response = self._post(path, raw_body, verify=True)
        return response

    def positions_history(self, **kwargs):
        """`Bitfinex positions history reference
        <https://docs.bitfinex.com/v2/reference#positions-history>`_

        Return the positions of a user between two dates

        Parameters
        ----------
        start : Optional int
            Millisecond start time

        end : Optional int
            Millisecond end time

        limit : Optional int
            Number of records

        Returns
        -------
        list
             ::

                [
                  [
                    SYMBOL,
                    STATUS,
                    AMOUNT,
                    BASE_PRICE,
                    MARGIN_FUNDING,
                    MARGIN_FUNDING_TYPE,
                    PL,
                    PL_PERC,
                    PRICE_LIQ,
                    LEVERAGE,
                    ID,
                    MTS_CREATE,
                    MTS_UPDATE
                  ],
                  ...
                ]

        Examples
        --------
         ::

            positions = bfx_client.positions_history(limit=10)
            for position in positions:
                print(position)

        """

        body = kwargs
        raw_body = json.dumps(body)
        path = "v2/auth/r/positions/hist"
        response = self._post(path, raw_body, verify=True)
        return response

    def positions_audit(self, **kwargs):
        """`Bitfinex positions audit reference
        <https://docs.bitfinex.com/v2/reference#positions-audit>`_

        Return and audit of the positions of a user that correspond to the ids send

        Parameters
        ----------
        id : Optional list of ints
            List of position IDs to audit

        start : Optional int
            Millisecond start time

        end : Optional int
            Millisecond end time

        limit : Optional int
            Number of records

        Returns
        -------
        list
             ::

                [
                  [
                    SYMBOL,
                    STATUS,
                    AMOUNT,
                    BASE_PRICE,
                    MARGIN_FUNDING,
                    MARGIN_FUNDING_TYPE,
                    PL,
                    PL_PERC,
                    PRICE_LIQ,
                    LEVERAGE,
                    ID,
                    MTS_CREATE,
                    MTS_UPDATE,
                    TYPE,
                    COLLATERAL,
                    COLLATERAL_MIN,
                    META
                  ],
                  ...
                ]

        Examples
        --------
         ::

            positions = bfx_client.positions_audit([1, 2, 3])
            for position in positions:
                print(position)

        """

        body = kwargs
        raw_body = json.dumps(body)
        path = "v2/auth/r/positions/audit"
        response = self._post(path, raw_body, verify=True)
        return response

    def funding_offers(self, symbol=""):
        """`Bitfinex funding offers reference
        <https://bitfinex.readme.io/v2/reference#rest-auth-funding-offers>`_

        Get active funding offers.

        Parameters
        ----------
        symbol : str
            The `symbol <restv1.html#symbols>`_ you want information about.

        Returns
        -------
        list
             ::

                [
                  [
                    ID,
                    SYMBOL,
                    MTS_CREATED,
                    MTS_UPDATED,
                    AMOUNT,
                    AMOUNT_ORIG,
                    TYPE,
                    _PLACEHOLDER,
                    _PLACEHOLDER,
                    FLAGS,
                    STATUS,
                    _PLACEHOLDER,
                    _PLACEHOLDER,
                    _PLACEHOLDER,
                    RATE,
                    PERIOD,
                    NOTIFY,
                    HIDDEN,
                    _PLACEHOLDER,
                    RENEW,
                    ...
                  ],
                  ...
                ]

        Examples
        --------
         ::

            bfx_client.funding_offers()

            bfx_client.funding_offers("fIOT")

        """
        body = {}
        raw_body = json.dumps(body)
        path = "v2/auth/r/funding/offers/{}".format(symbol)
        response = self._post(path, raw_body, verify=True)
        return response

    def funding_offers_history(self, symbol="", **kwargs):
        """`Bitfinex funding offers hist reference
        <https://bitfinex.readme.io/v2/reference#rest-auth-funding-offers-hist>`_

        Get past inactive funding offers. Limited to last 3 days.

        Parameters
        ----------
        symbol : str
            The `symbol <restv1.html#symbols>`_ you want information about.

        start : Optional int
            Millisecond start time

        end : Optional int
            Millisecond end time

        limit : Optional int
            Number of records

        Returns
        -------
        list
             ::

                [
                  [
                    ID,
                    SYMBOL,
                    MTS_CREATED,
                    MTS_UPDATED,
                    AMOUNT,
                    AMOUNT_ORIG,
                    TYPE,
                    _PLACEHOLDER,
                    _PLACEHOLDER,
                    FLAGS,
                    STATUS,
                    _PLACEHOLDER,
                    _PLACEHOLDER,
                    _PLACEHOLDER,
                    RATE,
                    PERIOD,
                    NOTIFY,
                    HIDDEN,
                    _PLACEHOLDER,
                    RENEW,
                    ...
                  ],
                  ...
                ]

        Examples
        --------
         ::

            bfx_client.funding_offers_history()

            bfx_client.funding_offers_history('fOMG')

        """
        body = kwargs
        raw_body = json.dumps(body)
        add_symbol = "{}/".format(symbol) if symbol else ""
        path = "v2/auth/r/funding/offers/{}hist".format(add_symbol)
        response = self._post(path, raw_body, verify=True)
        return response

    def funding_loans(self, symbol=""):
        """`Bitfinex funding loans reference
        <https://bitfinex.readme.io/v2/reference#rest-auth-funding-loans>`_

        Funds not used in active positions

        Parameters
        ----------
        symbol : str
            The `symbol <restv1.html#symbols>`_ you want information about.

        Returns
        -------
        list
             ::

                [
                  [
                    ID,
                    SYMBOL,
                    SIDE,
                    MTS_CREATE,
                    MTS_UPDATE,
                    AMOUNT,
                    FLAGS,
                    STATUS,
                    _PLACEHOLDER,
                    _PLACEHOLDER,
                    _PLACEHOLDER,
                    RATE,
                    PERIOD,
                    MTS_OPENING,
                    MTS_LAST_PAYOUT,
                    NOTIFY,
                    HIDDEN,
                    _PLACEHOLDER,
                    RENEW,
                    _PLACEHOLDER,
                    NO_CLOSE,
                    ...
                  ],
                  ...
                ]

        Example
        -------
         ::

            bfx_client.funding_loans('fOMG')

        """
        body = {}
        raw_body = json.dumps(body)
        path = "v2/auth/r/funding/loans/{}".format(symbol)
        response = self._post(path, raw_body, verify=True)
        return response

    def funding_loans_history(self, symbol="", **kwargs):
        """`Bitfinex funding loans history reference
        <https://bitfinex.readme.io/v2/reference#rest-auth-funding-loans-hist>`_

        Inactive funds not used in positions. Limited to last 3 days.

        Parameters
        ----------
        symbol : str
            The `symbol <restv1.html#symbols>`_ you want information about.

        start : Optional int
            Millisecond start time

        end : Optional int
            Millisecond end time

        limit : Optional int
            Number of records

        Returns
        -------
        list
             ::

                [
                  [
                    ID,
                    SYMBOL,
                    SIDE,
                    MTS_CREATE,
                    MTS_UPDATE,
                    AMOUNT,
                    FLAGS,
                    STATUS,
                    _PLACEHOLDER,
                    _PLACEHOLDER,
                    _PLACEHOLDER,
                    RATE,
                    PERIOD,
                    MTS_OPENING,
                    MTS_LAST_PAYOUT,
                    NOTIFY,
                    HIDDEN,
                    _PLACEHOLDER,
                    RENEW,
                    _PLACEHOLDER,
                    NO_CLOSE,
                    ...
                  ],
                  ...
                ]

        Example
        -------
         ::

            bfx_client.funding_loans_history('fOMG')

        """
        body = kwargs
        raw_body = json.dumps(body)
        add_symbol = "{}/".format(symbol) if symbol else ""
        path = "v2/auth/r/funding/loans/{}hist".format(add_symbol)
        response = self._post(path, raw_body, verify=True)
        return response

    def funding_credits(self, symbol=""):
        """`Bitfinex funding credits reference
        <https://bitfinex.readme.io/v2/reference#rest-auth-funding-credits>`_

        Funds used in active positions

        Parameters
        ----------
        symbol : str
            The `symbol <restv1.html#symbols>`_ you want information about.

        Returns
        -------
        list
             ::

                [
                  [
                    ID,
                    SYMBOL,
                    SIDE,
                    MTS_CREATE,
                    MTS_UPDATE,
                    AMOUNT,
                    FLAGS,
                    STATUS,
                    _PLACEHOLDER,
                    _PLACEHOLDER,
                    _PLACEHOLDER,
                    RATE,
                    PERIOD,
                    MTS_OPENING,
                    MTS_LAST_PAYOUT,
                    NOTIFY,
                    HIDDEN,
                    _PLACEHOLDER,
                    RENEW,
                    _PLACEHOLDER,
                    NO_CLOSE,
                    ...
                  ],
                  ...
                ]

        Example
        -------
         ::

            bfx_client.funding_credits('fUSD')

        """
        body = {}
        raw_body = json.dumps(body)
        path = "v2/auth/r/funding/credits/{}".format(symbol)
        response = self._post(path, raw_body, verify=True)
        return response

    def funding_credits_history(self, symbol="", **kwargs):
        """`Bitfinex funding credits history reference
        <https://bitfinex.readme.io/v2/reference#rest-auth-funding-credits-hist>`_

        Inactive funds used in positions. Limited to last 3 days.

        Parameters
        ----------
        symbol : str
            The `symbol <restv1.html#symbols>`_ you want information about.

        start : Optional int
            Millisecond start time

        end : Optional int
            Millisecond end time

        limit : Optional int
            Number of records

        Returns
        -------
        list
             ::

                [
                  [
                    ID,
                    SYMBOL,
                    SYMBOL,
                    MTS_CREATE,
                    MTS_UPDATE,
                    AMOUNT,
                    FLAGS,
                    STATUS,
                    _PLACEHOLDER,
                    _PLACEHOLDER,
                    _PLACEHOLDER,
                    RATE,
                    PERIOD,
                    MTS_OPENING,
                    MTS_LAST_PAYOUT,
                    NOTIFY,
                    HIDDEN,
                    _PLACEHOLDER,
                    RENEW,
                    _PLACEHOLDER,
                    NO_CLOSE,
                    POSITION_PAIR,
                    ...
                  ],
                  ...
                ]

        Example
        -------
         ::

            bfx_client.funding_credits_history('fUSD')

        """
        body = kwargs
        raw_body = json.dumps(body)
        add_symbol = "{}/".format(symbol) if symbol else ""
        path = "v2/auth/r/funding/credits/{}hist".format(add_symbol)
        response = self._post(path, raw_body, verify=True)
        return response

    def funding_trades(self, symbol="", **kwargs):
        """`Bitfinex funding trades hitory reference
        <https://bitfinex.readme.io/v2/reference#rest-auth-funding-trades-hist>`_

        Get funding trades

        Parameters
        ----------
        symbol : str
            The `symbol <restv1.html#symbols>`_ you want information about.

        start : Optional int
            Millisecond start time

        end : Optional int
            Millisecond end time

        limit : Optional int
            Number of records

        Returns
        -------
        list
             ::

                [
                  [
                    ID,
                    CURRENCY,
                    MTS_CREATE,
                    OFFER_ID,
                    AMOUNT,
                    RATE,
                    PERIOD,
                    MAKER,
                    ...
                  ],
                  ...
                ]

        Example
        -------
         ::

            bfx_client.funding_trades('fUSD')

        """
        body = kwargs
        raw_body = json.dumps(body)
        add_symbol = "{}/".format(symbol) if symbol else ""
        path = "v2/auth/r/funding/trades/{}hist".format(add_symbol)
        response = self._post(path, raw_body, verify=True)
        return response

    def margin_info(self, tradepair="base"):
        """`Bitfinex margin info reference
        <https://bitfinex.readme.io/v2/reference#rest-auth-info-margin>`_

        Get account margin info

        Parameters
        ----------
        key : str
            "base" | SYMBOL

        Returns
        -------
        list
             ::

                # margin base
                [
                  "base",
                  [
                    USER_PL,
                    USER_SWAPS,
                    MARGIN_BALANCE,
                    MARGIN_NET,
                    ...
                  ]
                ]

                # margin symbol
                [
                  SYMBOL,
                  [
                    TRADABLE_BALANCE,
                    GROSS_BALANCE,
                    BUY,
                    SELL,
                    ...
                  ]
                ]

        Examples
        --------
         ::

            bfx_client.margin_info()

            bfx_client.margin_info('base')

            bfx_client.margin_info('tIOTUSD')

        """
        body = {}
        raw_body = json.dumps(body)
        path = "v2/auth/r/info/margin/{}".format(tradepair)
        response = self._post(path, raw_body, verify=True)
        return response

    def funding_info(self, tradepair):
        """`Bitfinex funding info reference
        <https://bitfinex.readme.io/v2/reference#rest-auth-info-funding>`_

        Get account funding info

        Parameters
        ----------
        symbol : str
            The `symbol <restv1.html#symbols>`_ you want information about.

        Returns
        -------
        list
             ::

                [
                  "sym",
                  SYMBOL,
                  [
                    YIELD_LOAN,
                    YIELD_LEND,
                    DURATION_LOAN,
                    DURATION_LEND,
                    ...
                  ],
                  ...
                ]

        Example
        -------
         ::

            bfx_client.funding_info('fIOT')

        """
        body = {}
        raw_body = json.dumps(body)
        path = "v2/auth/r/info/funding/{}".format(tradepair)
        response = self._post(path, raw_body, verify=True)
        return response

    def movements(self, currency="", **kwargs):
        """`Bitfinex movements reference
        <https://bitfinex.readme.io/v2/reference#movements>`_

        View your past deposits/withdrawals.

        Parameters
        ----------
        currency : str
            Currency (BTC, ...)

        start : Optional int
            Millisecond start time

        end : Optional int
            Millisecond end time

        limit : Optional int
            Number of records, default & max: 25

        Returns
        -------
        list
             ::

                [
                  [
                    ID,
                    CURRENCY,
                    CURRENCY_NAME,
                    null,
                    null,
                    MTS_STARTED,
                    MTS_UPDATED,
                    null,
                    null,
                    STATUS,
                    null,
                    null,
                    AMOUNT,
                    FEES,
                    null,
                    null,
                    DESTINATION_ADDRESS,
                    null,
                    null,
                    null,
                    TRANSACTION_ID,
                    null
                  ],
                  ...
                ]

        Example
        -------
         ::

            bfx_client.movements()

            bfx_client.movements("BTC")

        """
        body = kwargs
        raw_body = json.dumps(body)
        add_currency = "{}/".format(currency.upper()) if currency else ""
        path = "v2/auth/r/movements/{}hist".format(add_currency)
        response = self._post(path, raw_body, verify=True)
        return response

    def performance(self, period="1D"):
        """`Bitfinex performance reference
        <https://bitfinex.readme.io/v2/reference#rest-auth-performance>`_

        Get account historical daily performance (work in progress on Bitfinex side)

        This endpoint is still under active development so you might experience unexpected behavior
        from it.

        Currently not working : bitfinex.rest.restv2.BitfinexException:
        (500, 'Internal Server Error', ['error', 10020, 'method: invalid'])

        Returns
        -------
        list
            The list contains the following information::

                [ CURRENT_RATE ]

        Example
        -------
         ::

            bfx_client.performance()

        """
        body = {}
        raw_body = json.dumps(body)
        path = "v2/auth/r/stats/perf:{}/hist".format(period)
        response = self._post(path, raw_body, verify=True)
        return response

    def alert_list(self):
        """`Bitfinex list alerts reference
        <https://docs.bitfinex.com/v2/reference#rest-auth-alert-list>`_

        List of active alerts

        Returns
        -------
        list
             ::

                [
                  [
                    'price:tBTCUSD:560.92',
                    'price',
                    'tBTCUSD',
                    560.92,
                    91
                  ],
                  ...
                ]

        Example
        -------
         ::

            bfx_client.alert_list()

        """
        body = {'type': 'price'}
        raw_body = json.dumps(body)
        path = "v2/auth/r/alerts"
        response = self._post(path, raw_body, verify=True)
        return response

    def alert_set(self, alert_type, symbol, price):
        """`Bitfinex auth alert set reference
        <https://bitfinex.readme.io/v2/reference#rest-auth-alert-set>`_

        Sets up a price alert at the given value

        Parameters
        ----------
        type : str
            Only one type is available : "price"

        symbol : str
            The `symbol <restv1.html#symbols>`_ you want information about.

        price : float
            Price where you want to receive alerts

        Returns
        -------
        list
             ::

                [
                  'price:tBTCUSD:600',
                  'price',
                  'tBTCUSD',
                  600,
                  100
                ]

        Example
        -------
         ::

            bfx_client.alert_set('price', 'tIOTUSD', 3)

        """
        body = {
            'type': alert_type,
            'symbol': symbol,
            'price': price
        }

        raw_body = json.dumps(body)
        path = "v2/auth/w/alert/set"
        response = self._post(path, raw_body, verify=True)
        return response

    def alert_delete(self, symbol, price):
        """`Bitfinex auth alert delete reference
        <https://bitfinex.readme.io/v2/reference#rest-auth-alert-delete>`_


        Bitfinex always returns [True] no matter if the request deleted an alert or not

        Parameters
        ----------
        symbol : str
            The `symbol <restv1.html#symbols>`_ you want information about.

        price : float
            Price where you want to receive alerts

        Returns
        -------
        list
             ::

                [ True ]

        Example
        -------
         ::

            bfx_client.alert_delete('tIOTUSD', 1)

        """
        body = {}

        raw_body = json.dumps(body)
        path = "v2/auth/w/alert/price:{}:{}/del".format(symbol, price)
        response = self._post(path, raw_body, verify=True)
        return response

    def calc_available_balance(self, symbol, direction, rate, order_type):
        """`Bitfinex calc balance available reference
        <https://bitfinex.readme.io/v2/reference#rest-auth-calc-bal-avail>`_

        Calculate available balance for order/offer
        example : calc_available_balance("tIOTUSD","1","1.13","EXCHANGE")

        Parameters
        ----------
        symbol : str
            The `symbol <restv1.html#symbols>`_ you want information about.

        dir : int
            direction of the order/offer
            (orders: > 0 buy, < 0 sell | offers: > 0 sell, < 0 buy)

        rate : string
            Rate of the order/offer

        type : string
            Type of the order/offer EXCHANGE or MARGIN

        Returns
        -------
        list
             ::

                [AMOUNT_AVAIL]

        Example
        -------
         ::

            bfx_client.calc_available_balance('tIOTUSD', 1, 0.02, 'EXCHANGE')

        """

        body = {
            'symbol': symbol,
            'dir': direction,
            'rate': rate,
            'type': order_type
        }

        raw_body = json.dumps(body)
        path = "v2/auth/calc/order/avail"
        response = self._post(path, raw_body, verify=True)
        return response

    def ledgers(self, currency=""):
        """`Bitfinex ledgers reference
        <https://bitfinex.readme.io/v2/reference#ledgers>`_

        View your past ledger entries.

        Parameters
        ----------
        Currency : str
            Currency (BTC, ...)

        Returns
        -------
        list
             ::

            [
              [
                ID,
                CURRENCY,
                null,
                TIMESTAMP_MILLI,
                null,
                AMOUNT,
                BALANCE,
                null,
                DESCRIPTION
              ],
              ...
            ]

        Example
        --------
         ::

            bfx_client.ledgers('IOT')

        """
        body = {}
        raw_body = json.dumps(body)
        add_currency = "{}/".format(currency.upper()) if currency else ""
        path = "v2/auth/r/ledgers/{}hist".format(add_currency)
        response = self._post(path, raw_body, verify=True)
        return response

    def user_settings_read(self, pkey):
        """`Bitfinex user settings read reference
        <https://bitfinex.readme.io/v2/reference#user-settings-read>`_

        Read user settings

        Parameters
        ----------
        pkey : str
            Requested Key

        Returns
        -------
        list
             ::

                [
                  [
                    KEY
                    VALUE
                  ],
                  ...
                ]

        Example
        --------
         ::

            none available

        """
        body = {
            'keys': ['api:{}'.format(pkey)]
        }
        raw_body = json.dumps(body)
        path = "v2/auth/r/settings"
        response = self._post(path, raw_body, verify=True)
        return response

    def user_settings_write(self, pkey):
        """`Bitfinex user settings write reference
        <https://bitfinex.readme.io/v2/reference#user-settings-write>`_

        Write user settings

        Warning
        -------
        Not Implemented

        """
        raise NotImplementedError

    def user_settings_delete(self, pkey):
        """`Bitfinex user settings delete reference
        <https://bitfinex.readme.io/v2/reference#user-settings-delete>`_

        Delete user settings

        Warning
        -------
        Not Implemented

        """
        raise NotImplementedError



import pandas as pd
import Quandl


class CSVDataSource:
    """
    Loads a CSV into a pandas dataframe
    """

    def __init__(self, fname, fields):
        """
        loads the CSV
        :param fname: csv file name
        :param fields: header names
        """
        self.data = pd.read_csv(fname, names=fields, parse_dates=True)

    def parse_timestamp_column(self, label, unit, set_index=True):
        """
        Convert (if necessary) a given column into a pandas timestamp type
        to allow handling of time series
        :param label: column to check
        :param unit: if column is already a timestamp, i.e. an integer, whats the time unit.
        :return:
        """
        if isinstance(self.data[label][0], pd.tslib.Timestamp):
            if set_index:
                self.data.set_index(label, inplace=True)
            return
        self.data[label] = pd.to_datetime(self.data[label], unit=unit)
        if set_index:
            self.data.set_index(label, inplace=True)



"""Bitfinex main module"""
from .rest.restv1 import Client
from .rest.restv2 import Client as ClientV2
from .websockets.client import WssClient

# This is for packward compatability.
ClientV1 = Client



"""Module for rest and websocket utilities"""
import re
import time
from datetime import datetime

def create_cid():
    """Create a new Client order id. Based on timestamp multiplied to 10k to
    make it improbable that two actions are assigned the same cid.

    Returns
    -------
    int
        A integer number equal to the current timestamp * 10 thousend.
    """
    now = datetime.utcnow()
    epoch = datetime.utcfromtimestamp(0)
    return int((now - epoch).total_seconds() * 10000)

def cid_to_date(cid):
    """Converts a cid to date string YYYY-MM-DD

    Parameters
    ----------
    cid : int
        A cid as it is generated by the function ``utils.create_cid()``

    Returns
    -------
    str
        A string formated date (e.g. YYYY-MM-DD, 2018-10-01)
    """
    return datetime.utcfromtimestamp(
        cid/10000.0
    ).strftime("%Y-%m-%d")

def get_nonce(multiplier):
    """Returns a nonce used in authentication.
    Nonce must be an increasing number. If other frameworks have used
    higher numbers you might need to increase the nonce_multiplier.
    """
    return str(float(time.time()) * multiplier)

TRADE_SYMBOL_MISSING = re.compile(r"^[a-zA-Z]{6}$")
"""Regular explression used to match trade symbols without a leading t (e.g. BTCUSD)"""

FUNDING_SYMBOL_MISSING = re.compile(r"^[a-zA-Z]{3}$")
"""Regular explression used to match funcing symbols without a leading f (e.g. BTC)"""

def order_symbol(symbol, capital=True):
    """Convinience function for skipping t or f before symbols for trade and
    funding orders.

    Parameters
    ----------
    symbol : str
        Symbol as a string. For example BTCUSD for trades or BTC for funding.

    capital : bool
        Boolean to capitalize trading and funding symbols. Capilat symbols are
        required in v2 of the Bitfinex API. Default: True
    """
    if capital:
        _symbol = symbol.upper()
    else:
        _symbol = symbol

    if TRADE_SYMBOL_MISSING.match(symbol):
        return "t{}".format(_symbol)
    elif FUNDING_SYMBOL_MISSING.match(symbol):
        return "f{}".format(_symbol)

    return symbol
